import json
import xml.etree.ElementTree as elementTree

import httpx
import validators
from returns.future import FutureFailure

from dino_seedwork_be.exceptions import MainException

validator_utils = validators


def is_email(a_value: str):
    return validators.email(a_value)


def is_in_json_format(aText: str):
    try:
        json.loads(aText)
        return True
    except ValueError:
        return False


def is_xml(value):
    try:
        elementTree.fromstring(value)
    except elementTree.ParseError:
        return False
    return True


def is_url_image(image_url):
    image_formats = ("image/png", "image/jpeg", "image/jpg")
    r = httpx.head(image_url)
    if r.headers["content-type"] in image_formats:
        return True
    return False


def kw_page_size_validate(function):
    def wrapper(*args, **kwargs):
        page = kwargs["page"]
        size = kwargs["size"]
        if page < 1:
            raise MainException(code="PAGE_INVALID_VALUE")
        if size <= 0:
            raise MainException(code="SIZE_INVALID_VALUE")
        return function(*args, **kwargs)

    return wrapper


def safe_kw_page_size_validate(function):
    def wrapper(*args, **kwargs):
        try:
            page = kwargs["page"]
            size = kwargs["size"]
            if page < 1:
                return FutureFailure(MainException(code="PAGE_INVALID_VALUE"))
            if size <= 0:
                return FutureFailure(MainException(code="SIZE_INVALID_VALUE"))
            return function(*args, **kwargs)
        except KeyError:
            return function(*args, **kwargs)

    return wrapper



from functools import reduce
from typing import Any, List, Optional

from sqlalchemy import TEXT, Column, and_, cast, func, or_
from sqlalchemy.exc import PendingRollbackError
from sqlalchemy.ext.asyncio.session import AsyncSession
from sqlalchemy.util import EMPTY_DICT

from dino_seedwork_be.adapters.persistance.sql.DBSessionUser import \
    DBSessionUser
from dino_seedwork_be.application.query import BaseQuerier
from dino_seedwork_be.exceptions import MainException
from dino_seedwork_be.implementation.adapter.rest import (FilterElement,
                                                          OrderParam,
                                                          ParamOperators,
                                                          ParamWithComparing)
from dino_seedwork_be.utils.list import remove_none

__all__ = ["AlchemyQuerier"]


class AlchemyQuerier(BaseQuerier):
    _query_session: AsyncSession

    def set_session(self, session: AsyncSession):
        self._query_session = session

    def apply_session(self, sessionUsers: List[DBSessionUser]):
        for user in sessionUsers:
            user.set_session(self.session())

    def session(self) -> AsyncSession:
        return self._query_session

    async def execute(
        self,
        statement,
        params=None,
        execution_options=EMPTY_DICT,
        bind_arguments=None,
        **kw
    ) -> Any:
        execute = lambda: self.session().execute(
            statement, params, execution_options, bind_arguments, **kw
        )
        try:
            result = await execute()
            return result
        except PendingRollbackError:
            await self.session().rollback()
            return await execute()

    def is_allowed_primitive(self, v: Any):
        return (
            isinstance(v, bool)
            or isinstance(v, str)
            or isinstance(v, int)
            or isinstance(v, float)
        )

    def a_param_to_colum_filter(
        self, param: Any | ParamWithComparing, column: Column, is_collection: bool
    ):
        match param:
            case ParamWithComparing():
                match param.operator:
                    case ParamOperators.LT:
                        return column < param.value
                    case ParamOperators.LTE:
                        return column <= param.value
                    case ParamOperators.GT:
                        return column > param.value
                    case ParamOperators.GTE:
                        return column >= param.value
                    case ParamOperators.EQ:
                        return column == param.value
                    case _:
                        raise MainException(code="UNKNOWN_PARAM_COMPARATOR")
            case p if self.is_allowed_primitive(p):
                match is_collection:
                    case False:
                        return column == param
                    case True:
                        return column.any(param)
            case _:
                return None

    def param_list_to_column_filter(
        self,
        param: List[Any] | List[ParamWithComparing],
        column: Column,
        is_collection: bool,
    ):
        match param:
            case [*elems] if all(isinstance(e, ParamWithComparing) for e in elems):
                return and_(
                    *[
                        self.a_param_to_colum_filter(p, column, is_collection)
                        for p in param
                    ]
                )
            case [*elems] if all(self.is_allowed_primitive(e) for e in elems):
                return or_(
                    *[
                        self.a_param_to_colum_filter(p, column, is_collection)
                        for p in param
                    ]
                )
            case _:
                return None

    def param_to_column_filter(
        self, param: FilterElement, column: Column, is_collection: bool = False
    ):
        match param:
            case list():
                return remove_none(
                    [self.param_list_to_column_filter(param, column, is_collection)]
                )
            case _:
                return remove_none(
                    [self.a_param_to_colum_filter(param, column, is_collection)]
                )

    @staticmethod
    def gen_space_interwine_columns(columns: List[Column]):
        def reduce_interwine_with_space(acc, value):
            match value[0]:
                case int() as index if index == len(columns) - 1:
                    return [*acc, value[1]]
                case _:
                    return [*acc, value[1], " "]

        return func.concat(*reduce(reduce_interwine_with_space, enumerate(columns), []))

    def fuzzy_search(self, columns: List[Column], q: Optional[str] = None):
        distance_score = (
            AlchemyQuerier.gen_space_interwine_columns(columns)
            .op("<->")(cast(q, TEXT))
            .label("distance_score")
        )
        match q:
            case str(text) if text.strip() != "":
                return {
                    "select": [distance_score],
                    "filter": [distance_score < 0.95],
                    "order": [distance_score],
                }
            case _:
                return {"select": [], "filter": [], "order": []}

    def get_order_query(self, column: Column, order: Optional[OrderParam]):
        match order:
            case OrderParam.ASC:
                return [column.asc()]
            case OrderParam.DESC:
                return [column.desc()]
            case None:
                return []

    def get_order_query_list(self, orders: List[tuple[Column, OrderParam]]):
        def order_to_db_query(order: tuple[Column, OrderParam]):
            match order[1]:
                case OrderParam.ASC:
                    return order[0].asc()
                case OrderParam.DESC:
                    return order[0].desc()

        return map(order_to_db_query, orders)



from abc import ABC, abstractmethod
from typing import Any, Generic, TypeVar

from returns.result import Result

SerializableType = TypeVar("SerializableType")


class AbstractSerializer(ABC, Generic[SerializableType]):
    @abstractmethod
    def serialize(self, obj: SerializableType) -> Result[str, Any]:
        pass

    @abstractmethod
    def deserialize(self, an_json: str) -> Result[SerializableType, Any]:
        pass



import json
from typing import Any, List

from dino_seedwork_be.implementation.adapter.rest.utils import to_param_orders
from dino_seedwork_be.utils.dict import extract

__all__ = ["Orders"]


class Orders:
    def __init__(self, keys: List[str]) -> None:
        self.keys = keys

    def __call__(self, orders: Any = {}) -> Any:
        match orders:
            case str():
                try:
                    orders = json.loads(orders)
                except Exception:
                    orders = {}

        plainOrders = extract(orders, self.keys)
        return to_param_orders(plainOrders)



from typing import List, Optional

from dino_seedwork_be.exceptions import MainException

# __all__ = [
#     "DomainException",
#     "BusinessRuleValidationException",
#     "DomainIllegalStateException",
#     "DomainIllegalArgumentException",
# ]


class DomainException(MainException):
    pass


class BusinessRuleValidationException(DomainException):
    pass


class DomainIllegalArgumentException(BusinessRuleValidationException):
    def __init__(
        self,
        message: str,
        loc: Optional[List[str]] = None,
        code: Optional[str] = None,
    ):
        super().__init__(
            message=f"IllegalArgumentException in Domain Logic: {message}",
            loc=loc,
            code=code,
        )


class DomainIllegalStateException(BusinessRuleValidationException):
    def __init__(
        self,
        message: str,
        loc: Optional[List[str]] = None,
        code: Optional[str] = None,
    ):
        super().__init__(
            message=f"IllegalStateException in Domain Logic: {message}",
            loc=loc,
            code=code,
        )



from abc import ABC, abstractmethod
from datetime import datetime
from typing import Callable, List, Optional

from pika.exchange_type import ExchangeType
from returns.functions import tap
from returns.future import FutureResult, FutureSuccess
from returns.maybe import Maybe
from returns.pipeline import flow
from returns.pointfree import bind, map_
from returns.result import Result, safe

from dino_seedwork_be.adapters.logger.SimpleLogger import DomainLogger
from dino_seedwork_be.adapters.messaging.notification.EventHandlingTracker import \
    EventHandlingTracker
from dino_seedwork_be.utils.functional import feed_args, tap_excute_future

from .ConnectionSettings import ConnectionSettings
from .Exchange import Exchange
from .MessageConsumer import MessageConsumer
from .MessageListener import MessageListener
from .MessageType import MessageType as Type
from .Queue import Queue

# __all__ = ["ExchangeListener"]


class ExchangeListener(ABC):

    _message_consumer: Optional[MessageConsumer]
    _queue: Optional[Queue] = None
    _exchange: Exchange

    is_exchange_durable: bool = True
    is_exchange_auto_delete: bool = False
    exchange_type: ExchangeType = ExchangeType.fanout

    is_queue_durable: bool = True
    is_queue_auto_deleted: bool = False
    is_queue_exclusive: bool = True
    queue_routing_keys: Optional[List[str]] = None

    is_auto_acknowledged: bool = False
    is_retry: bool = False

    label: str = ""
    logger: DomainLogger
    event_handling_tracker: EventHandlingTracker

    def _queue_routing_keys(self) -> List[str]:
        return Maybe.from_optional(self.queue_routing_keys).value_or(self.listen_to())

    def __init__(self, connection_setting: ConnectionSettings) -> None:
        self._connection_setting = connection_setting
        self.attach_to_queue().unwrap()
        self.logger = DomainLogger(self.label)

    def message_consumer(self) -> Maybe[MessageConsumer]:
        return Maybe.from_optional(self._message_consumer)

    @safe
    def close(self):
        """
        " Closes my queues
        """
        self.queue().map(lambda queue: queue.close())

    def queue(self) -> Maybe[Queue]:
        return Maybe.from_optional(self._queue)

    def exchange(self) -> Exchange:
        return self._exchange

    @abstractmethod
    def exchange_name(self) -> str:
        """
        " Answers the String name of the exchange I listen to.
        " @return str
        """
        pass

    @abstractmethod
    def filtered_dispatch(self, a_type: str, a_text_message: str) -> FutureResult:
        ...

    @abstractmethod
    def listen_to(self) -> List[str]:
        """
        " Answers the kinds of messages I listen to.
        " :return: List[str]
        """
        pass

    @abstractmethod
    def queue_name(self) -> str:
        """
        " Answers the str name of the queue I listen to. By
        " default it is the simple name of my concrete class.
        " May be overridden to change the name.
        " :return: str
        """
        pass

    def set_exchange(self, exchange: Exchange):
        self._exchange = exchange

    def set_queue(self, a_queue: Queue):
        self._queue = a_queue

    def attach_to_queue(self) -> Result:

        """
        " Attaches to the queues I listen to for messages.
        """
        return flow(
            [
                self._connection_setting,
                self.exchange_name(),
                self.exchange_type,
                self.is_exchange_durable,
                self.is_exchange_auto_delete,
                lambda exchange: flow(
                    [
                        exchange,
                        self.queue_name(),
                        self._queue_routing_keys(),
                        self.is_queue_durable,
                        self.is_queue_exclusive,
                        self.is_queue_auto_deleted,
                        self.register_consumer,
                    ],
                    feed_args(Queue.factory_exchange_subscriber),
                    map_(tap(self.set_queue)),
                ),
            ],
            feed_args(Exchange.factory),
            map_(tap(self.set_exchange)),
        )

    def set_message_consumer(self, message_consumer: MessageConsumer):
        self._message_consumer = message_consumer

    def itempotent_handle_dispatch(
        self, a_message_id: str, a_type: str, a_binary_message: bytes
    ):
        def idempotent_handle(is_handled: bool):
            match is_handled:
                case False:
                    return self.filtered_dispatch(
                        a_type, a_binary_message.decode()
                    ).bind(
                        tap_excute_future(
                            lambda _: self.event_handling_tracker.mark_notif_as_handled(
                                a_message_id=a_message_id
                            )
                        )
                    )
                case _:
                    return FutureSuccess(None)

        self.logger.info(f"Handle message_id {a_message_id}")
        return self.event_handling_tracker.check_if_notif_handled(a_message_id).bind(
            idempotent_handle
        )

    def register_consumer(self, queue: Queue) -> Result:
        parent = self

        class MessageListenerAdapter(MessageListener):
            def handle_message(
                self,
                a_type: str,
                a_message_id: str,
                a_time_stamp: datetime,
                a_binary_message: bytes,
                a_delivery_tag: int,
                is_redelivery: bool,
            ) -> FutureResult:
                return parent.itempotent_handle_dispatch(
                    a_message_id, a_type, a_binary_message
                )

        rabbitmq_register_consumer: Callable[
            [MessageConsumer], Result
        ] = lambda msg_consumer: msg_consumer.receive_only(
            self.listen_to(), MessageListenerAdapter(Type.TEXT)
        )

        return flow(
            [self.is_auto_acknowledged, queue, self.is_retry, self.label],
            feed_args(MessageConsumer.factory),
            map_(tap(self.set_message_consumer)),
            bind(rabbitmq_register_consumer),
        )

    def is_ready_for_consume(self) -> bool:
        return self.message_consumer().map(lambda ms: ms.is_consuming()).value_or(False)

    def run(self):
        self.exchange().run()

    def stop(self):
        self.message_consumer().map(tap(lambda ms: ms.close()))



from abc import abstractmethod
from dataclasses import dataclass
from typing import Any, Generic, List, TypeVar

from returns.future import FutureResult
from returns.maybe import Maybe

from dino_seedwork_be.domain.Entity import Entity
from dino_seedwork_be.domain.value_object.AbstractIdentity import \
    AbstractIdentity
from dino_seedwork_be.exceptions import NotImplementError

EntityType = TypeVar("EntityType", bound=Entity)
DTOType = TypeVar("DTOType")


@dataclass
class PaginationResultDB(Generic[EntityType]):
    items: List[EntityType]
    limit: int
    size: int
    total: int


class Repository(Generic[EntityType]):
    @abstractmethod
    def get_next_id(self, simple: bool = False) -> FutureResult[AbstractIdentity, Any]:
        pass

    @abstractmethod
    def get_by_id(self, id: AbstractIdentity) -> FutureResult[Maybe[EntityType], Any]:

        raise NotImplementError(
            "Repository does not support the default getById"
            "method, you need to give it the detail logic"
        )

    @abstractmethod
    def add(self, entity: EntityType) -> FutureResult:

        raise NotImplementError(
            "Repository does not support the default insert"
            "method, you need to give it the detail logic"
        )

    @abstractmethod
    def save(self, entity: EntityType) -> FutureResult:

        raise NotImplementError(
            "Repository does not support the default update"
            "method, you need to give it the detail logic"
        )

    @abstractmethod
    def remove(self, id: AbstractIdentity) -> FutureResult:
        raise NotImplementError(
            "Repository does not support the default delete"
            "method, you need to give it the detail logic"
        )

    @abstractmethod
    def count(self) -> FutureResult[int, Any]:
        raise NotImplementError(
            "Repository does not support the default count"
            "method, you need to give it the detail logic"
        )



from abc import abstractmethod
from typing import Any

from returns.future import FutureResult
from returns.maybe import Maybe

from dino_seedwork_be.adapters.messaging.notification.NotificationPublisher import \
    NotificationPublisher
from dino_seedwork_be.application.ApplicationLifeCycleUseCase import \
    ApplicationLifeCycleUsecase

__all__ = ["DefaultImplementNotificationApplicationService"]


class DefaultImplementNotificationApplicationService(ApplicationLifeCycleUsecase):

    _notification_publisher: NotificationPublisher

    def __init__(self, notification_publisher: NotificationPublisher) -> None:
        self._notification_publisher = notification_publisher

    def notification_publisher(self) -> NotificationPublisher:
        return self._notification_publisher

    @abstractmethod
    def publish_notifications(self) -> FutureResult:
        ...

    def _publish_notifications(self) -> FutureResult[Maybe[int], Any]:
        """
        publish unpublished event in event store
        :return: return the last event int id that published already
        """
        return self.notification_publisher().publish_notifications()



from typing import Generic

from dino_seedwork_be.domain.IdentifiedDomainObject import IdentityType

from .Entity import Entity, RawAttributes

# __all__ = ["AggregateRoot"]


class AggregateRoot(
    Generic[RawAttributes, IdentityType], Entity[RawAttributes, IdentityType]
):
    """Consits of 1+ entities. Spans transaction boundaries."""

    ...



from .AbstractApplicationLifeCycle import AbstractApplicationServiceLifeCycle
from .ApplicationLifeCycleUseCase import ApplicationLifeCycleUsecase
from .command import Command
from .handler import BaseHandler
from .query import BaseQuerier, PaginationResult
from .service import AbstractUOWApplicationService, UowFactory

__all__ = [
    "AbstractApplicationServiceLifeCycle",
    "ApplicationLifeCycleUsecase",
    "Command",
    "BaseHandler",
    "BaseQuerier",
    "PaginationResult",
    "AbstractUOWApplicationService",
    "UowFactory",
]



import json
import re
from dataclasses import dataclass
from datetime import datetime
from enum import Enum
from functools import reduce
from typing import (Any, Dict, Generic, List, Literal, Optional, TypedDict,
                    TypeVar)

from pydantic import BaseModel
from returns.maybe import Maybe, Nothing, Some
from toolz.dicttoolz import assoc

from dino_seedwork_be.domain.value_object.AbstractValueObject import \
    ValueObject
from dino_seedwork_be.exceptions import MainException
from dino_seedwork_be.utils.text import parse_num_or_keeping

DataType = TypeVar("DataType")

# __all__ = [
#     "RestPayload",
#     "to_rest_payload",
#     "pagination",
#     "field_error_validation",
#     "error_validation",
#     "error_validation_standard",
#     "error_detail_with_code_standard",
#     "error_detail_with_code",
#     "FilterWithPag",
#     "ParamOperator",
#     "ParamOperators",
#     "ParamWithComparing",
#     "FilterElement",
#     "PlainFilterSet",
#     "FilterSet",
#     "Filter",
#     "OrderParam",
#     "plain_order_to_param_order",
#     "to_param_orders",
# ]


class RestPayload(Generic[DataType], BaseModel):
    data: DataType
    _link: dict


def to_rest_payload(data: object | str = "OK", link: dict = {}):
    return {"data": data, "_link": link}


def pagination(total: int, offset: int, limit: int, items: list):
    return {
        "items": items,
        "offset": offset,
        "limit": limit,
        "total": total,
    }


def field_error_validation(loc: List[str], msg: str, type: Optional[str]):
    return {"loc": loc, msg: str, type: type}


def error_validation(detail: dict | List):
    return {"validation": detail}


def error_validation_standard(exception: MainException):
    return {
        "validation": [
            {
                "loc": exception.loc().value_or(None),
                "code": exception.code().value_or(None),
                "message": exception.message().value_or(None),
            }
        ]
    }


def error_detail_with_code_standard(exception: MainException):
    return {
        "origin": "DINO_MARKET_BACKEND",
        "code": exception.code(),
        "message": exception.message(),
        "loc": exception.loc(),
    }


def error_detail_with_code(exception: MainException, is_validation: bool = False):
    message = {}
    try:
        message = json.loads(exception.message().value_or("{}"))
    except Exception:
        pass
    return {
        "origin": "DINO_MARKET_BACKEND",
        "code": exception.code(),
        "validation" if is_validation else "message": message,
    }


EntityAttrsType = TypeVar("EntityAttrsType", bound=TypedDict)


class FilterWithPag(Generic[EntityAttrsType]):
    filters: EntityAttrsType
    order: Dict[str, Literal["ASC", "DESC"]]
    offset: int
    limit: int


@dataclass
class ParamOperator(ValueObject):
    symbol: str

    def __eq__(self, __o: object) -> bool:
        match __o:
            case ParamOperator():
                return self.symbol == __o.symbol
            case _:
                return False

    def __hash__(self) -> int:
        return hash(self.symbol)

    def is_lt(self):
        return self.symbol == "lt"

    def is_lte(self):
        return self.symbol == "lte"

    def is_gt(self):
        return self.symbol == "gt"

    def is_gte(self):
        return self.symbol == "gte"

    def is_eq(self):
        return self.symbol == "eq"


class ParamOperators(Enum):
    LT = ParamOperator("lt")
    LTE = ParamOperator("lte")
    GT = ParamOperator("gt")
    GTE = ParamOperator("gte")
    EQ = ParamOperator("eq")


@dataclass
class ParamWithComparing:
    operator: ParamOperators
    value: str | int | float | datetime

    def __eq__(self, __o: object) -> bool:
        if isinstance(__o, ParamWithComparing):
            return __o.operator == self.operator and __o.value == self.value
        return False


FilterElement = Any | ParamWithComparing | List[Any] | List[ParamWithComparing]

PlainFilterSet = Dict[str, Any]

FilterSet = Dict[str, FilterElement]


class Filter:
    parsed_filter: FilterSet
    filters: PlainFilterSet
    order: Dict[str, Literal["ASC", "DESC"]]
    q: Optional[str]

    def __init__(self, plainFilter: Dict[str, Any]) -> None:
        self.filters = plainFilter
        self.parsed_filter = self.parse_filters()
        super().__init__()

    def parse_filter(self, param_value: Any):
        def parse_a_instance(param_elem: Any):
            pattern = "~(.*?)~"
            pop_opr = (re.search(pattern, param_elem) or re.Match()).group(1)
            pop_no_str = re.sub(r"^~.*?~", "", param_elem)
            return ParamWithComparing(
                operator=ParamOperators(ParamOperator(pop_opr)),
                value=parse_num_or_keeping(pop_no_str),
            )

        try:
            match param_value:
                case list():
                    return list(map(lambda v: parse_a_instance(v), param_value))
                case _:
                    return parse_a_instance(param_value)
        except Exception as error:
            print("cannot parse ", error)
            return param_value

    def parse_filters(self):
        return reduce(
            lambda acc, key: assoc(acc, key, self.parse_filter(self.filters[key])),
            self.filters.keys(),
            {},
        )


class OrderParam(Enum):
    ASC = "ASC"
    DESC = "DESC"


def plain_order_to_param_order(plainOrder: Optional[str]) -> Maybe[OrderParam]:
    match plainOrder:
        case "asc":
            return Some(OrderParam.ASC)
        case "desc":
            return Some(OrderParam.DESC)
        case _:
            return Nothing


def to_param_orders(plainOrders: dict[str, Optional[str]]):
    return reduce(
        lambda acc, key: assoc(
            acc, key, plain_order_to_param_order(plainOrders[key]).value_or(None)
        ),
        plainOrders.keys(),
        {},
    )



import threading
from typing import Generic, TypeVar

InnerValue = TypeVar("InnerValue")


class ThreadLocal(Generic[InnerValue]):
    _storage: threading.local
    _key: str

    def __init__(self, key: str, initValue: InnerValue):
        self._storage = threading.local()
        self._key = key
        setattr(self._storage, key, initValue)

    def value(self) -> InnerValue:
        return getattr(self._storage, self._key)

    def set_value(self, aValue: InnerValue):
        setattr(self._storage, self._key, aValue)



from typing import Any

from returns.maybe import Some
from returns.result import Result

from dino_seedwork_be.domain.value_object.AbstractIdentity import \
    AbstractIdentity

# __all__ = ["NID"]


class NID(AbstractIdentity[int]):
    def validate(self, an_id: int) -> Result[int, Any]:
        return self.assert_argument_larger_than(
            an_id, 0, code=Some("NUMBER_ID_SHOULD_>_0")
        )



def none_or_instance(cls, param, *args):
    if param is None:
        return None
    return cls(param, *args)


def none_or_transform(obj, transfomer):
    if obj is None:
        return None
    return transfomer(obj)



from typing import List

from returns.maybe import Maybe, Nothing, Some

from dino_seedwork_be.domain.value_object.AbstractValueObject import \
    ValueObject

# __all__ = ["URL"]


class URL(ValueObject):
    _value: str
    _regex: str = r"https?:\/\/(www\.)?[-a-zA-Z0-9@:%._\+~#=]{2,256}\.[a-z]{2,4}\b([-a-zA-Z0-9@:%_\+.~#?&//=]*)"

    def __eq__(self, other):
        if isinstance(other, URL):
            return self.getValue() == other.getValue()
        return False

    def __init__(
        self,
        value: str,
        validation_message: Maybe[str] = Nothing,
        loc: List[str] = ["url"],
    ):
        self.assert_argument_regex(
            value, self._regex, a_message=validation_message, loc=Some(loc)
        ).unwrap()
        self._value = value
        super().__init__()

    def getValue(self):
        return self._value



from typing import Any, Callable, Optional

from pika.exchange_type import ExchangeType
from returns.maybe import Maybe
from returns.result import Result, safe

from dino_seedwork_be.adapters.logger.SimpleLogger import SIMPLE_LOGGER
from dino_seedwork_be.utils.functional import execute

from .BrokerComponent import BrokerComponent
from .ConnectionSettings import ConnectionSettings

# __all__ = ["Exchange"]


class Exchange(BrokerComponent):
    _type: ExchangeType
    _is_exchange_ready: bool = False
    _is_auto_delete: bool = False

    def is_exchange(self) -> bool:
        return True

    def set_exchange_ready_status(self, a_bool: bool):
        self._is_exchange_ready = a_bool

    def is_exchange_ready(self) -> bool:
        return self._is_exchange_ready

    def __init__(
        self,
        a_con_settings: ConnectionSettings,
        a_name: str,
        a_type: ExchangeType,
        is_durable: bool,
        is_auto_delete: bool = False,
        on_setup_finish: Optional[Callable[..., Result]] = None,
    ) -> None:
        super().__init__(a_name, a_con_settings, None, on_setup_finish)
        self.set_type(a_type)
        self.set_durable(is_durable)
        self._is_auto_delete = is_auto_delete

    @staticmethod
    @safe
    def factory(
        a_con_settings: ConnectionSettings,
        a_name: str,
        a_type: ExchangeType,
        is_durable: bool,
        is_auto_delete: bool = False,
        on_setup_finish: Optional[Callable[..., Result]] = None,
    ):
        return Exchange(
            a_con_settings, a_name, a_type, is_durable, is_auto_delete, on_setup_finish
        )

    @staticmethod
    @safe
    def direct_instance(
        a_con_settings: ConnectionSettings,
        a_name: str,
        is_durable: bool,
        is_auto_delete: bool = False,
        on_setup_finish: Optional[Callable[..., Result]] = None,
    ) -> "Exchange":
        return Exchange(
            a_con_settings,
            a_name,
            ExchangeType.direct,
            is_durable,
            is_auto_delete,
            on_setup_finish=on_setup_finish,
        )

    @staticmethod
    @safe
    def fanout_instance(
        a_con_settings,
        a_name: str,
        is_durable: bool,
        is_auto_delete: bool = False,
        on_setup_finish: Optional[Callable[..., Result]] = None,
    ) -> "Exchange":
        return Exchange(
            a_con_settings,
            a_name,
            ExchangeType.fanout,
            is_auto_delete,
            is_durable,
            on_setup_finish=on_setup_finish,
        )

    @staticmethod
    @safe
    def headers_instance(
        a_con_settings,
        a_name: str,
        is_durable: bool,
        is_auto_delete: bool = False,
        on_setup_finish: Optional[Callable[..., Result]] = None,
    ) -> "Exchange":
        return Exchange(
            a_con_settings,
            a_name,
            ExchangeType.headers,
            is_durable,
            is_auto_delete,
            on_setup_finish=on_setup_finish,
        )

    @staticmethod
    @safe
    def topic_instance(
        a_con_settings,
        a_name: str,
        is_durable: bool,
        is_auto_delete: bool = False,
        on_setup_finish: Optional[Callable[..., Result]] = None,
    ) -> "Exchange":
        return Exchange(
            a_con_settings,
            a_name,
            ExchangeType.topic,
            is_durable,
            is_auto_delete,
            on_setup_finish=on_setup_finish,
        )

    def type(self) -> ExchangeType:
        return self._type

    def set_type(self, a_type: ExchangeType):
        self._type = a_type

    @safe
    def setup(self, callback: Optional[Callable[[Any], Result]] = None):
        SIMPLE_LOGGER.info("setup an exchange %s", self.name())

        def on_exchange_declare_ok(_):
            SIMPLE_LOGGER.info("delare exchange %s successfully :)", self.name())
            self.set_exchange_ready_status(True)
            Maybe.from_optional(execute(callback, self)).map(
                lambda result: result.unwrap()
            )

        self.channel().unwrap().exchange_declare(
            exchange=self.name(),
            exchange_type=self.type(),
            durable=self.is_durable(),
            auto_delete=self._is_auto_delete,
            callback=on_exchange_declare_ok,
        )



from typing import Optional

from validators.btc_address import re

# __all__ = ["not_empty", "test_regex"]


def not_empty(a_value: str, a_message: Optional[str] = None):
    if a_value.strip() == "":
        raise ValueError(a_message or "Value cannot be empty")


def test_regex(a_value: str, a_regex: str, a_message: Optional[str] = None):
    if re.match(a_regex, a_value) is None:
        raise ValueError(a_message or "Value is not in valid format")



from typing import Callable, List, TypeVar

from returns.future import FutureResult, future_safe
from returns.pipeline import flow
from returns.pointfree import bind, lash
from sqlalchemy.ext.asyncio.session import AsyncSession

from dino_seedwork_be.adapters.persistance.sql.AbstractUnitOfWork import \
    AbstractUnitOfWork
from dino_seedwork_be.adapters.persistance.sql.DBSessionUser import \
    DBSessionUser
from dino_seedwork_be.utils.functional import pass_to

ResultType = TypeVar("ResultType")
ExceptionType = TypeVar("ExceptionType", bound=Exception)

__all__ = ["SqlAlchemyUnitOfWork"]


class SqlAlchemyUnitOfWork(AbstractUnitOfWork[AsyncSession]):

    _session: AsyncSession

    def __init__(
        self,
        session_factory: Callable[[], AsyncSession],
        session_users: List[DBSessionUser[AsyncSession]],
    ):
        super().__init__(session_users, session_factory)

    def session(self) -> AsyncSession:
        return self._session

    def set_session(self, session: AsyncSession):
        self._session = session

    async def __aenter__(
        self,
    ):
        self.set_session(self.session_factory()())
        [
            session_user.set_session(self._session)
            for session_user in self.session_users()
        ]
        result = await super().__aenter__()
        return result

    @future_safe
    async def future_Enter(self):
        r = await self.__aenter__()
        return r

    async def __aexit__(self, *args):
        await super().__aexit__(*args)
        await self._session.close()

    @future_safe
    async def future__aexit_with_failure(self, result: ExceptionType) -> ExceptionType:
        await self.__aexit__(result)
        print("result failure", result)
        raise result

    @future_safe
    async def future__aexit__with_success(self, result: ResultType) -> ResultType:
        await self.__aexit__(None)
        return result

    async def _commit(self):
        await self._session.commit()
        print("commit to db ", self.session_users)

    async def rollback(self):
        await self._session.rollback()
        print("rollback db ", self.session_users)

    def absord(self):
        def enter(_):
            return self.future_Enter()

        def catch_future_result(
            result: FutureResult[ResultType, ExceptionType]
        ) -> FutureResult[ResultType, ExceptionType]:
            return flow(
                True,
                enter,
                bind(pass_to(result)),
                bind(self.future__aexit__with_success),
                lash(self.future__aexit_with_failure),
            )

        return catch_future_result

    @classmethod
    def factory(
        cls,
        session_factory: Callable[[], AsyncSession],
        session_users: List[DBSessionUser],
    ):
        return SqlAlchemyUnitOfWork(session_factory, session_users)



from functools import reduce
from typing import Generic, Iterable, List, TypeVar


class DuplicateKeyError(Exception):
    pass


ItemType = TypeVar("ItemType")


class ValidateSet(set[ItemType], Generic[ItemType]):
    def __init__(self, val: Iterable[ItemType]):
        super().__init__(val)

    def add(self, value):
        if value in self:
            raise DuplicateKeyError("Value {!r} already present".format(value))
        super().add(value)

    def update(self, values):
        error_values = []
        for value in values:
            if value in self:
                error_values.append(value)
        if error_values:
            raise DuplicateKeyError(
                "Value(s) {!r} already present".format(error_values)
            )
        super().update(values)

    def copy(self):
        return ValidateSet(super().copy())


def set_add(s, x):
    return len(s) != (s.add(x) or len(s))


def set_remove(s, x):
    try:
        s.remove(x)
        return True
    except KeyError:
        return False


def set_from(listItem: List):
    def reduce_func(acc: ValidateSet, item):
        new_set = acc.copy()
        new_set.add(item)
        return new_set

    return reduce(reduce_func, listItem, ValidateSet([]))



from abc import abstractmethod
from dataclasses import dataclass
from typing import Any, Generic, List, TypeVar

from returns.future import FutureResult

ItemType = TypeVar("ItemType")

# __all__ = ["PaginationResult", "BaseQuerier"]

InputType = TypeVar("InputType")
ReturnType = TypeVar("ReturnType")


@dataclass
class PaginationResult(Generic[ItemType]):
    items: List[ItemType]
    total: int
    offset: int
    limit: int


class BaseQuerier:
    pass


class BusBaseQuerier(Generic[InputType, ReturnType], BaseQuerier):
    @abstractmethod
    def handle(self, query: InputType) -> FutureResult[ReturnType, Any]:
        pass

    def execute(self, query: InputType) -> FutureResult[ReturnType, Any]:
        return self.handle(query)



from abc import abstractmethod
from typing import Generic, TypeVar

from returns.result import Result

from dino_seedwork_be.domain.exceptions import DomainException

RepositoryModel = TypeVar("RepositoryModel")
AggregateRootModel = TypeVar("AggregateRootModel")


class Mapper(Generic[AggregateRootModel, RepositoryModel]):
    @classmethod
    @abstractmethod
    def to_aggregate(
        cls,
        a_db_model: RepositoryModel,
    ) -> Result[AggregateRootModel, DomainException]:
        pass

    @classmethod
    @abstractmethod
    def to_db(
        cls, an_aggregate: AggregateRootModel
    ) -> Result[RepositoryModel, DomainException]:
        pass



import logging
from typing import Any, Callable, List, Optional, Union

from multimethod import multimethod
from returns.curry import partial
from returns.iterables import Fold
from returns.maybe import Maybe
from returns.result import Failure, Result, Success, safe

from dino_seedwork_be.adapters.logger.SimpleLogger import SIMPLE_LOGGER
from dino_seedwork_be.exceptions import MainException
from dino_seedwork_be.utils import execute, unsafe_panic

from .BrokerComponent import BrokerComponent
from .ConnectionSettings import ConnectionSettings
from .Exchange import Exchange

LOGGER = logging.getLogger(__name__)

# __all__ = ["Queue"]


class Queue(BrokerComponent):

    _is_auto_deleted: bool
    _is_exclusive: bool
    _is_queue_ready: bool = False

    def is_queue(self) -> bool:
        return True

    def __init__(
        self,
        a_con_settings: Optional[ConnectionSettings],
        a_broker_component: Optional[Union[Exchange, "Queue"]],
        a_name: str,
        is_durable: bool,
        is_exclusive: bool,
        is_auto_deleted: bool,
        on_setup_finish: Optional[Callable[[Any], Any]],
    ):
        """
        Construct my default state.
        @param a_con_settings the ConnectionSettings to initialize with
        @param a_broker_channel the BrokerChannel to initialize with
        @param a_name the String name of the queue, or the empty string
        @param is_durable the boolean indicating whether or not I am durable
        @param is_exclusive the boolean indicating whether or not I am exclusive
        @param is_auto_deleted the boolean indicating whether or not I should be auto-deleted
        """
        self._is_exclusive = is_exclusive
        self._is_auto_deleted = is_auto_deleted
        self.set_durable(is_durable)
        super().__init__(a_name, a_con_settings, a_broker_component, on_setup_finish)

    @multimethod
    @staticmethod
    @safe
    def factory(
        a_con_settings: ConnectionSettings,
        a_name: str,
        on_setup_finish: Optional[Callable[[Any], Any]],
    ):
        """
        Answers a new instance of a Queue with the name a_name. The underlying
        queue is non-durable, non-exclusive, and not auto-deleted.
        @param a_connection_settings the ConnectionSettings
        @param a_name the String name of the queue
        @return Queue
        """
        return Queue(
            a_con_settings,
            None,
            a_name,
            False,
            False,
            False,
            on_setup_finish=on_setup_finish,
        )

    @safe
    @factory.register
    @staticmethod
    def _(
        a_con_settings: ConnectionSettings,
        a_name: str,
        is_durable: bool,
        is_exclusive: bool,
        is_auto_deleted: bool,
        on_setup_finish: Optional[Callable[[Any], Any]],
    ) -> "Queue":
        """
        Answers a new instance of a Queue with the name a_name. The underlying
        queue durability, exclusivity, and deletion properties are specified by
        explicit parameters.
        @param a_connection_settings the ConnectionSettings
        @param a_name the String name of the queue
        @param is_durable the boolean indicating whether or not I am durable
        @param is_exclusive the boolean indicating whether or not I am exclusive
        @param is_auto_deleted the boolean indicating whether or not I should be auto-deleted
        @return Queue
        """
        return Queue(
            a_con_settings,
            None,
            a_name,
            is_durable,
            is_exclusive,
            is_auto_deleted,
            on_setup_finish,
        )

    @staticmethod
    @safe
    def factory_durable_instance(
        a_connection_settings: ConnectionSettings,
        a_name: str,
        on_setup_finish: Optional[Callable[[Any], Any]],
    ) -> "Queue":
        """
        Answers a new instance of a Queue with the name a_name. The underlying
        queue is durable, exclusive, and not auto-deleted.
        @param a_connection_settings the ConnectionSettings
        @param a_name the String name of the queue
        @return Queue
        """

        return Queue(
            a_connection_settings,
            None,
            a_name,
            True,
            False,
            False,
            on_setup_finish,
        )

    @staticmethod
    @safe
    def factory_durable_exclusive_instance(
        a_connection_settings: ConnectionSettings,
        a_name: str,
        on_setup_finish: Optional[Callable],
    ) -> "Queue":
        """
        * Answers a new instance of a Queue with the name aName. The underlying
        * queue is durable, exclusive, and not auto-deleted.
        * @param a_connection_settings the ConnectionSettings
        * @param a_name the String name of the queue
        * @return Queue
        """
        return Queue(
            a_connection_settings,
            None,
            a_name,
            True,
            True,
            False,
            on_setup_finish,
        )

    @multimethod
    @staticmethod
    @safe
    def factory_exchange_subscriber(
        an_exchange: Exchange,
        a_name: str,
        cb: Optional[Callable[[Any], Result]] = None,
        is_durable: bool = False,
        is_auto_deleted: bool = True,
        is_exclusive: bool = True,
    ):
        SIMPLE_LOGGER.info(
            "Factory a QUEUE is durable %s, is_auto_deleted %s, is_exclusive %s"
            % (is_durable, is_auto_deleted, is_exclusive)
        )

        def callback(_):
            return Queue._bind_queue(queue, an_exchange, cb, "")

        queue = Queue(
            None,
            an_exchange,
            a_name or "",
            is_durable,
            is_exclusive,
            is_auto_deleted,
            callback,
        )
        return queue

    @staticmethod
    @factory_exchange_subscriber.register
    @safe
    def _(
        an_exchange: Exchange,
        routing_keys: List[str],
        cb: Optional[Callable] = None,
        is_durable: bool = False,
        is_auto_deleted: bool = True,
        is_exclusive: bool = True,
    ):
        """
        Answers a new instance of a Queue that is bound to an_exchange, and
        is ready to participate as an exchange subscriber (pub/sub). The
        connection and channel of an_exchange are reused. The Queue is
        uniquely named by the server, non-durable, exclusive, and auto-deleted.
        The queue is bound to all routing keys in routing_keys. This Queue
        style best works as a temporary direct or topic subscriber.
        @param an_exchange the Exchange to bind with the new Queue
        @return Queue
        """

        def callback(_):
            return Fold.collect(
                map(partial(Queue._bind_queue, queue, an_exchange, cb), routing_keys),
                Result(()),
            ).unwrap()

        queue = Queue(
            None,
            an_exchange,
            "",
            is_durable,
            is_exclusive,
            is_auto_deleted,
            callback,
        )
        return queue

    @staticmethod
    @factory_exchange_subscriber.register
    @safe
    def _(
        an_exchange: Exchange,
        a_name: str,
        routing_keys: List[str],
        is_durable: bool,
        is_exclusive: bool,
        is_auto_deleted: bool,
        cb: Optional[Callable],
    ) -> "Queue":
        """
        Answers a new instance of a Queue that is bound to an_exchange, and
        is ready to participate as an exchange subscriber (pub/sub). The
        connection and channel of an_exchange are reused. The Queue is named
        by aName, unless it is empty, in which case the name is generated by
        the broker. The Queue is bound to all routing keys in routing_keys,
        or to no routing key if aRoutingKeys is empty. The Queue has the
        qualities specified by is_durable, is_exclusive, is_auto_deleted. This
        factory is provided for ultimate flexibility in case no other
        exchange-queue binder factories fit the needs of the client.
        @param an_exchange the Exchange to bind with the new Queue
        @param a_name the String name of the queue
        @param routing_keys the routing keys to bind the queue to
        @param is_durable the boolean indicating whether or not I am durable
        @param is_exclusive the boolean indicating whether or not I am exclusive
        @param is_auto_deleted the boolean indicating whether or not I should be auto-deleted
        @return Queue
        """

        def callback(_):
            match len(routing_keys):
                case 0:
                    return Queue._bind_queue(queue, an_exchange, cb, "")
                case _:
                    return Fold.collect(
                        map(
                            partial(Queue._bind_queue, queue, an_exchange, cb),
                            routing_keys,
                        ),
                        Success(()),
                    ) or Result(None)

        queue = Queue(
            None,
            an_exchange,
            a_name,
            is_durable,
            is_exclusive,
            is_auto_deleted,
            on_setup_finish=callback,
        )
        return queue

    @staticmethod
    def _bind_queue(
        a_queue: "Queue",
        an_exchange: Exchange,
        cb: Optional[Callable[[Any], Result]],
        routing_key: str,
    ) -> Result:
        def on_bind_ok(_):
            LOGGER.info("Queue bind successfully")
            execute_result = execute(cb, a_queue)
            Maybe.from_optional(execute_result).map(lambda r: r.unwrap())
            a_queue.set_queue_ready_status(True)

        try:
            a_queue.channel().unwrap().queue_bind(
                a_queue.name(), an_exchange.name(), routing_key, callback=on_bind_ok
            )
            return Success(None)
        except Exception as error:
            return Failure(
                MainException(code="BIND_QUEUE_TO_EXCHANGE_FAILED", message=str(error))
            )

    @safe
    def setup(self, callback: Optional[Callable[[Any], Result]]):
        try:
            self.channel().unwrap().queue_declare(
                queue=self.name(),
                durable=self.is_durable(),
                exclusive=self._is_exclusive,
                auto_delete=self._is_auto_deleted,
                callback=lambda _: unsafe_panic(callback)(self)
                if callback is not None
                else None,
            )
        except Exception as error:
            raise MainException(code="QUEUE_OPEN_FAILED", message=str(error))

    def is_queue_ready(self):
        return self._is_queue_ready

    def set_queue_ready_status(self, is_ready: bool):
        self._is_queue_ready = is_ready



from typing import Any, Dict, Type

from returns.maybe import Maybe, Nothing, Some

from dino_seedwork_be.domain.DomainEvent import DomainEvent

# __all__ = ["AbstractDomainEventDict"]


class AbstractDomainEventDict:
    _dict: Dict[str, Any] = {}

    # panic
    @classmethod
    def register(cls, name: str, type: Type[DomainEvent]):
        # cls.get_type_by_key(name).map(
        #     lambda _: throwException(
        #         DomainException(
        #             code="TYPE_EXISTED_IN_DOMAIN_MODEL_DICT",
        #             message=f"Type {name} already existed {type}",
        #         )
        #     )
        # )
        cls._dict[name] = type

    @classmethod
    def get_key_by_type(cls, type: Type) -> Maybe[str]:
        for key, value in cls._dict.items():
            match value == type:
                case True:
                    return Some(key)
        return Nothing

    @classmethod
    def get_type_by_key(cls, key: str) -> Maybe[Type[DomainEvent]]:
        try:
            return Some(cls._dict[key])
        except KeyError:
            return Nothing



from datetime import datetime

from returns.maybe import Maybe, Nothing

from dino_seedwork_be.utils.date import now_utc

from ..DomainAssertionConcern import DomainAssertionConcern

# __all__ = ["ValueObject"]


class ValueObject(DomainAssertionConcern):
    """
    Base class for value objects
    """

    _created_at: datetime

    def __init__(self, created_at: Maybe[datetime] = Nothing) -> None:
        self._created_at = created_at.value_or(now_utc())
        super().__init__()



from datetime import datetime
from typing import Any, Generic, TypeVar

from returns.maybe import Some
from returns.pipeline import flow
from returns.pointfree import bind
from returns.result import Failure, Result, Success, safe

from dino_seedwork_be.domain.DomainEvent import DomainEvent
from dino_seedwork_be.logic.assertion_concern import AssertionConcern
from dino_seedwork_be.serializer.Serializable import JSONSerializable
from dino_seedwork_be.utils.functional import return_v

__all__ = ["DomainEventSerializable", "Notification"]


class DomainEventSerializable(DomainEvent, JSONSerializable):
    pass


DomainEventT = TypeVar("DomainEventT", bound=DomainEvent)


class Notification(Generic[DomainEventT], JSONSerializable, AssertionConcern):
    _event: DomainEventT
    _id: int
    _occurred_on: datetime
    _type_name: str
    _version: int = 0

    def __init__(self, a_num_id: int, event: DomainEventT):
        self.set_id(a_num_id)
        self.set_domain_event(event)
        self.set_occurred_on(event.occurred_on())
        self.set_type_name(event.type())
        self.set_version(event.version())

    @safe
    @staticmethod
    def factory(a_num_id: int, event: DomainEventT):
        return Notification(a_num_id, event)

    def set_domain_event(self, event: DomainEventT):
        self.assert_argument_not_null(event, a_message=Some("Event must be specified"))
        self._event = event

    def set_id(self, a_num_id: int):
        self.assert_argument_not_null(
            a_num_id, a_message=Some("Notification id cannot be null")
        )
        self._id = a_num_id

    def set_version(self, a_version: int):
        self._version = a_version

    def set_type_name(self, a_type_name: str) -> Result[None, Any]:
        assert_result = flow(
            self.assert_argument_not_empty(
                Some(a_type_name), a_message=Some("The type name cannot be empty")
            ),
            bind(
                return_v(
                    self.assert_argument_length(
                        a_type_name,
                        a_maximum=100,
                        a_minimum=0,
                        a_message=Some("The type name must be 100 characters or less"),
                    )
                )
            ),
        )
        match assert_result:
            case Failure(_):
                return assert_result
            case _:
                self._type_name = a_type_name
                return Success(None)

    def set_occurred_on(self, anOccurredOn: datetime):
        self._occurred_on = anOccurredOn

    def type_name(self) -> str:
        return self._type_name

    def version(self) -> int:
        return self._version

    def occurred_on(self) -> datetime:
        return self._occurred_on

    def id(self) -> int:
        return self._id

    def event(self) -> DomainEvent:
        return self._event

    @staticmethod
    def restore(a_dict: dict) -> "Notification":
        event = a_dict["event"]
        return Notification(a_num_id=a_dict["id"], event=DomainEvent.restore(event))

    def as_dict(self) -> dict:
        return {"id": self.id(), "event": self.event().as_dict()}



import json
from datetime import datetime
from decimal import Decimal
from typing import Generic, TypeVar

from returns.maybe import Maybe

from dino_seedwork_be.media.AbstractJsonMediaReader import \
    AbstractJSONMediaReader

ValueType = TypeVar("ValueType")
DomainEventT = TypeVar("DomainEventT")
"""
    Reader for Message-Based Context Mapping 
"""

__all__ = ["NotificationReader"]


class NotificationReader(Generic[DomainEventT], AbstractJSONMediaReader):
    _event: dict

    def __init__(self, a_json_notification: str):
        notification_dict = json.loads(a_json_notification)

        self.set_event(notification_dict["event"])

    def event(self):
        return self._event

    def set_event(self, event: dict):
        self._event = event

    def event_string_value(self, path: str):
        return Maybe.from_optional(self.string_value(a_dict=self.event(), path=path))

    def event_big_decimal_value(self, path: str):
        str_val = self.string_value(a_dict=self.event(), path=path)
        return Maybe.from_optional(Decimal(str_val) if str_val is not None else None)

    def event_boolean_value(self, path: str) -> Maybe[bool]:
        return Maybe.from_optional(self.boolean_value(a_dict=self.event(), path=path))

    def event_datetime_value(self, path: str) -> Maybe[datetime]:
        return Maybe.from_optional(self.datetime_value(a_dict=self.event(), path=path))

    def event_float_value(self, path: str) -> Maybe[float]:
        return Maybe.from_optional(self.float_value(a_dict=self.event(), path=path))

    def event_int_value(self, path: str) -> Maybe[int]:
        return Maybe.from_optional(self.int_value(a_dict=self.event(), path=path))



import validators
from returns.maybe import Some

from dino_seedwork_be.domain.value_object.AbstractValueObject import \
    ValueObject
from dino_seedwork_be.utils.validator import is_url_image

# __all__ = ["ImageURL"]


class ImageURL(ValueObject):
    _url: str

    def __init__(self, url: str, validate_url: bool = False):
        self.set_url(url, validate_url)
        super().__init__()

    def set_url(self, url: str, validate_url: bool = False):
        self.assert_argument_not_empty(
            Some(url), a_message=Some("url string cannot be None")
        ).unwrap()
        validators.url(url)
        if validate_url:
            self.assert_state_true(
                is_url_image(url), Some("url is not a valid url image")
            ).unwrap()
        self._url = url

    def url(self) -> str:
        return self._url



import json
from typing import Optional

from returns.maybe import Maybe, Some
from returns.pipeline import pipe
from returns.result import safe

from dino_seedwork_be.domain.DomainEvent import DomainEvent
from dino_seedwork_be.serializer.AbstractSerializer import AbstractSerializer

# __all__ = ["EventSerializer"]


class EventSerializer(AbstractSerializer):
    ins: Optional["EventSerializer"] = None

    @classmethod
    def instance(cls) -> "EventSerializer":
        return (
            Maybe.from_optional(EventSerializer.ins)
            .lash(pipe(lambda _: cls.init_instance(), Some))
            .unwrap()
        )

    @classmethod
    def init_instance(cls):
        cls.ins = EventSerializer()
        return cls.ins

    @safe
    def serialize(self, an_event: DomainEvent) -> str:
        return str(json.dumps(an_event.as_dict(), default=str))

    @safe
    def deserialize(self, an_json: str) -> DomainEvent:
        return DomainEvent.restore(json.loads(an_json))



from abc import abstractmethod
from typing import Any, List

from firebase_admin.messaging import Notification
from returns.future import FutureResult
from returns.maybe import Maybe

from dino_seedwork_be.adapters.messaging.notification.PublishedNotificationTracker import \
    PublishedNotificationTracker

__all__ = ["PublishedNotificationTrackerStore"]


class PublishedNotificationTrackerStore:
    @abstractmethod
    def published_notification_tracker() -> FutureResult[
        PublishedNotificationTracker, Exception
    ]:
        ...

    @abstractmethod
    def track_most_recent_published_notification(
        self, a_tracker: PublishedNotificationTracker, notifications: List[Notification]
    ) -> FutureResult[Maybe[int], Any]:

        """
        persisting last event identifier that published in tracker
        :param a_tracker: tracker
        :notifications: List of published event as notifications
        :return: return the last event int id that published already
        """
        pass

    @abstractmethod
    def topic_name() -> str:
        ...



from abc import ABC, abstractmethod
from typing import Generic, List, TypeVar

from dino_seedwork_be.adapters.persistance.sql.DBSessionUser import \
    DBSessionUser

# __all__ = ["ApplicationLifeCycleUsecase"]

SessionType = TypeVar("SessionType")


class ApplicationLifeCycleUsecase(ABC, Generic[SessionType]):
    _session: SessionType

    def session(self):
        return self._session

    def set_session(self, a_session):
        self._session = a_session

    @abstractmethod
    def get_session_users(self) -> List[DBSessionUser[SessionType]]:
        ...



from abc import ABC, abstractmethod
from traceback import print_exception
from typing import Callable, Dict, Generic, List, Optional
from uuid import uuid4

from returns.functions import tap
from returns.future import FutureResult
from returns.pipeline import flow, managed
from returns.pointfree import alt, bind, lash
from returns.result import safe
from toolz.dicttoolz import dissoc

from dino_seedwork_be.adapters.persistance.sql.DBSessionUser import (
    DBSessionUser, SessionType)
from dino_seedwork_be.domain.event.EventStore import EventStore
from dino_seedwork_be.utils.functional import (apply, tap_excute_future,
                                               tap_failure_execute_future)

from .ApplicationLifeCycleUseCase import ApplicationLifeCycleUsecase

# __all__ = ["AbstractApplicationServiceLifeCycle"]


class AbstractApplicationServiceLifeCycle(ABC, Generic[SessionType]):
    _sessions: Dict = {}
    _event_store: EventStore

    @classmethod
    def event_store(cls) -> EventStore:
        return cls._event_store

    @classmethod
    def sessions(cls):
        return cls._sessions

    @classmethod
    @abstractmethod
    def session_factory(cls):
        pass

    @classmethod
    def add_session(cls, correlation_id: Optional[str] = None):
        correlation_id = correlation_id or str(uuid4())
        cls._sessions[correlation_id] = cls.session_factory()
        return [correlation_id, cls._sessions[correlation_id]]

    @classmethod
    def get_session_by_correlation(cls, correlation_id: str):
        session = cls._sessions[correlation_id]
        match session:
            case None:
                return cls.add_session(correlation_id)[1]
            case _:
                return session

    @classmethod
    @safe
    def remove_session_by_correlation(cls, correlation_id):
        cls._sessions = dissoc(cls._sessions, correlation_id)

    @classmethod
    @abstractmethod
    def initialize(cls):
        ...

    @classmethod
    def get_new_correlation_id(cls) -> str:
        return str(uuid4())

    @classmethod
    def begin(
        cls,
        correlation_id: str,
        db_session_users: List[DBSessionUser[SessionType]],
        usecase: ApplicationLifeCycleUsecase[SessionType],
    ) -> FutureResult:
        return cls.start_db(correlation_id, db_session_users, usecase).bind(
            lambda _: cls.event_listen(correlation_id)
        )

    @classmethod
    def exit(cls, correlation_id, result: FutureResult) -> FutureResult:
        commit_or_rollback_result = flow(
            result,
            bind(tap_excute_future(apply(cls.commit_db, correlation_id))),
            alt(tap(print_exception)),
            lash(tap_failure_execute_future(apply(cls.rollback_db, correlation_id))),
        )
        release_session = lambda *_: FutureResult.from_result(
            cls.remove_session_by_correlation(correlation_id)
        )
        return flow(
            commit_or_rollback_result,
            managed(lambda _: commit_or_rollback_result, release_session),
        )

    @classmethod
    @abstractmethod
    def rollback_db(cls, correlation_id: str) -> FutureResult:
        ...

    @classmethod
    @abstractmethod
    def commit_db(cls, correlation_id: str) -> FutureResult:
        ...

    @classmethod
    @abstractmethod
    def start_db(
        cls,
        correlation_id: str,
        db_session_users: List[DBSessionUser[SessionType]],
        usecase: ApplicationLifeCycleUsecase[SessionType],
    ) -> FutureResult:
        ...

    @classmethod
    @abstractmethod
    def event_listen(cls, correlation_id: str) -> FutureResult:
        ...

    @classmethod
    def life_cycle(cls):
        def decor(target_function: Callable[..., FutureResult]):
            def proxy(*args, **kwargs):
                db_session_users: List[DBSessionUser[SessionType]] = args[
                    0
                ].get_session_users()
                correlation_id = cls.get_new_correlation_id()
                return flow(
                    cls.begin(correlation_id, db_session_users, args[0]),
                    managed(
                        lambda _: target_function(*args, **kwargs),
                        lambda _, result: cls.exit(correlation_id, result),
                    ),
                )

            return proxy

        return decor



import json
from typing import Any, List, Tuple, TypeVar

from returns.curry import partial
from returns.future import FutureFailure, FutureResult, FutureSuccess
from returns.iterables import Fold
from returns.pipeline import flow, managed

from dino_seedwork_be.adapters.logger.SimpleLogger import DomainLogger
from dino_seedwork_be.exceptions import MainException
from dino_seedwork_be.utils import ThreadLocal

from .DomainEvent import DomainEvent
from .DomainEventSubscriber import DomainEventSubscriber

EventType = TypeVar("EventType", bound=DomainEvent)

# __all__ = ["DomainEventPublisher"]


class DomainEventPublisher:
    ins: ThreadLocal["DomainEventPublisher"]
    _isLock: bool
    _subscribers: List[DomainEventSubscriber]
    logger: DomainLogger = DomainLogger("DomainEventPublisher")

    def __init__(self) -> None:
        self._subscribers = []
        self._isLock = False

    @classmethod
    def instance(cls) -> "DomainEventPublisher":
        try:
            return DomainEventPublisher.ins.value()
        except AttributeError as error:
            cls.logger.info(f"attribute error in publisher {error}")
            DomainEventPublisher.new_instance_for_publisher()
            return DomainEventPublisher.ins.value()

    def is_lock(self) -> bool:
        return self._isLock

    def lock(self):
        self._isLock = True

    def unlock(self):
        self._isLock = False

    def has_subscribers(self) -> bool:
        match len(self._subscribers):
            case 0:
                return False
            case _:
                return True

    def publish(self, an_event: DomainEvent) -> FutureResult[Tuple, Any]:
        def check_if_target_event(event_type: str, subscriber: DomainEventSubscriber):
            match subscriber.event_type_subscribed():
                case list(target_event_types):
                    return event_type in target_event_types
                case str(target_event_type):
                    return target_event_type == event_type or target_event_type == "*"
                case _:
                    return False

        def execute(is_allow_to_run: bool):
            match is_allow_to_run:
                case True:
                    self.lock()
                    return flow(
                        self._subscribers,
                        partial(
                            filter,
                            partial(check_if_target_event, an_event.type()),
                        ),
                        partial(map, lambda sub: sub.handle_event(an_event)),
                        lambda futureS: Fold.collect(futureS, FutureSuccess(())),
                    )
                case False:
                    return FutureFailure(MainException(code="EVENT_PUBLISHER_LOCK"))

        self.logger.info(
            f"Publish domain event {json.dumps(an_event.as_dict())}, {self.is_lock()}, {self.has_subscribers()}",
        )
        return flow(
            FutureSuccess((not self.is_lock()) and self.has_subscribers()),
            managed(execute, lambda *_: FutureResult.from_value(self.unlock())),
        )

    def is_processing(self):
        return self.is_lock()

    def publish_all(self, events: List[DomainEvent]) -> FutureResult[Tuple, Any]:
        return flow(
            events,
            partial(map, lambda event: self.publish(event)),
            lambda futureS: Fold.collect(futureS, FutureSuccess(())),
        )

    def reset(self):
        match self.is_lock():
            case False:
                self.set_subscribers([])

    def set_subscribers(self, subs: List[DomainEventSubscriber]):
        self._subscribers = subs

    def subscribe(self, sub: DomainEventSubscriber):
        match self.is_lock():
            case False:
                self._subscribers.append(sub)

    @staticmethod
    def new_instance_for_publisher():
        DomainEventPublisher.ins = ThreadLocal(
            "domain_publisher", DomainEventPublisher()
        )


# DomainEventPublisher.new_instance_for_publisher()



import os

from dotenv import load_dotenv
from returns.curry import partial


def cast_bool_from_str(value):
    if value.lower() in ["true", "yes", "on", "1"]:
        value = True
    elif value.lower() in ["false", "no", "not", "off", "0"]:
        value = False
    else:
        raise ValueError(
            f'Incorrect value: "{value}". '
            f"It should be one of [1, 0, true, false, yes, no]"
        )
    return value


def get_env(name, default=None, is_bool=False, env="local"):
    load_dotenv(f".{env}.env")
    value = os.environ.get(name)
    if value is not None:
        if is_bool:
            return cast_bool_from_str(value)
        else:
            return value
    return default


def get_environment() -> str:
    return os.environ.get("ENV", default="local")


def get_env_with(env: str):
    return partial(get_env, env=env)



from typing import List, Type

from returns.future import future_safe
from returns.maybe import Maybe, Nothing, Some
from sqlalchemy import Column, Integer, String, select, update

from dino_seedwork_be.adapters.messaging.notification.Notification import \
    Notification
from dino_seedwork_be.adapters.messaging.notification.PublishedNotificationTracker import \
    PublishedNotificationTracker
from dino_seedwork_be.adapters.messaging.notification.PublishedNotificationTrackerStore import \
    PublishedNotificationTrackerStore
from dino_seedwork_be.implementation.adapter.storage.sql.alchemy.Repository import \
    StandardAlchemyRepository

__all__ = [
    "SqlAlchemyBasePublishedNotifTracker",
    "SqlAlchemyPublishedNotificationTrackerStore",
]


class SqlAlchemyBasePublishedNotifTracker:
    last_published_event_id = Column(Integer())
    type_name = Column(String(), primary_key=True, nullable=False)


class SqlAlchemyPublishedNotificationTrackerStore(
    PublishedNotificationTrackerStore, StandardAlchemyRepository
):
    _db_model: Type[SqlAlchemyBasePublishedNotifTracker]
    _type_name: str

    def __init__(
        self,
        db_model: Type[SqlAlchemyBasePublishedNotifTracker],
        type_name: str,
    ) -> None:
        self._db_model = db_model
        self._type_name = type_name
        super().__init__()

    @future_safe
    async def published_notification_tracker(self) -> PublishedNotificationTracker:
        stmt = select(self.db_model()).where(
            self.db_model().type_name == self.type_name()
        )
        tracker = (await self.session().execute(stmt)).scalars().first()
        match tracker:
            case None:
                new_tracker = PublishedNotificationTracker.factory(
                    type_name=self._type_name
                ).unwrap()
                self.session().add(self.db_model()(type_name=self._type_name))
                return new_tracker
            case _:
                return PublishedNotificationTracker.factory(
                    type_name=self.type_name(),
                    most_recent_published_notification_id=tracker.last_published_event_id,
                ).unwrap()

    @future_safe
    async def track_most_recent_published_notification(
        self, a_tracker: PublishedNotificationTracker, notifications: List[Notification]
    ) -> Maybe[int]:
        match len(notifications):
            case 0:
                return Nothing
            case _:
                last_notfi = notifications[-1]
                a_tracker.set_most_recent_published_notification_id(last_notfi.id())
                last_published_notif_id = (
                    a_tracker.most_recent_published_notification_id().value_or(None)
                )
                match last_published_notif_id:
                    case int(an_id):
                        stmt = (
                            update(self.db_model())
                            .where(self.db_model().type_name == a_tracker.type_name())
                            .values(last_published_event_id=an_id)
                        )
                        await self.session().execute(stmt)
                await self.session().commit()
                return Some(last_notfi.id())

    def db_model(self) -> Type[SqlAlchemyBasePublishedNotifTracker]:
        return self._db_model

    def type_name(self) -> str:
        return self._type_name



from typing import Any, Optional

from returns.maybe import Maybe, Some
from returns.pipeline import flow
from returns.pointfree import bind, map_
from returns.result import Failure, Result, Success

from dino_seedwork_be.logic.assertion_concern import AssertionConcern
from dino_seedwork_be.utils.functional import apply, return_v

__all__ = ["PublishedNotificationTracker"]


class PublishedNotificationTracker(AssertionConcern):
    _concurrency_version: int = 0
    _most_recent_published_notification_id: Optional[int] = None
    _id: Optional[int]
    _type_name: str

    def __init__(self):
        super().__init__()

    @staticmethod
    def factory(
        type_name: str,
        id: Optional[int] = None,
        most_recent_published_notification_id: Optional[int] = None,
    ) -> Result["PublishedNotificationTracker", Any]:
        tracker = PublishedNotificationTracker()
        return flow(
            tracker.set_type_name(type_name),
            map_(apply(tracker.set_id, id)),
            bind(
                apply(
                    tracker.set_most_recent_published_notification_id,
                    most_recent_published_notification_id,
                )
            ),
            map_(return_v(tracker)),
        )

    def concurrency_version(self) -> int:
        return self._concurrency_version

    def id(self) -> Maybe[int]:
        return Maybe.from_optional(self._id)

    def type_name(self) -> str:
        return self._type_name

    def most_recent_published_notification_id(self) -> Maybe[int]:
        return Maybe.from_optional(self._most_recent_published_notification_id)

    def set_most_recent_published_notification_id(
        self, an_num_id: int | None
    ) -> Result[None, Any]:
        match an_num_id:
            case int():
                assert_result = self.assert_argument_larger_than(
                    an_num_id, 0, Some("notification id must be large than zero")
                )
                match assert_result:
                    case Failure(_):
                        return assert_result
                    case _:
                        self._most_recent_published_notification_id = an_num_id
                        return Success(None)
        return Success(None)

    def set_type_name(self, an_type_name: str) -> Result[None, Any]:
        assert_condition = flow(
            self.assert_argument_not_empty(
                Some(an_type_name),
                code=Some("TRACKER_type_name_NOT_EMPTY"),
                a_message=Some("The tracker topic name is required"),
            ),
            bind(
                lambda _: self.assert_argument_length(
                    an_type_name,
                    1,
                    100,
                    Some("The tracker topic name must be 100 characters or less"),
                )
            ),
        )
        match assert_condition:
            case Success(_):
                self._type_name = an_type_name
                return Success(None)
            case _:
                return assert_condition

    def set_id(self, an_id: int):
        self._id = an_id

    def set_concurrency_version(self, version: int):
        self._concurrency_version = version



from typing import Any
from uuid import UUID, uuid4

from returns.result import Failure, Result, Success, safe

from dino_seedwork_be.domain.value_object.AbstractIdentity import \
    AbstractIdentity


class ProcessId(AbstractIdentity[str]):
    def __init__(self, id: str) -> None:
        super().__init__(id)

    def __eq__(self, __o: object) -> bool:
        match __o:
            case ProcessId():
                return __o.id() == self.id()
            case _:
                return False

    @staticmethod
    def new_process_id() -> Result["ProcessId", Any]:
        try:
            return Success(ProcessId(str(uuid4())))
        except Exception as error:
            return Failure(error)

    @staticmethod
    def from_existing_id(a_str: str) -> "ProcessId":
        return ProcessId(a_str)

    @safe
    def validate(self, an_id: str):
        UUID(an_id)
        return an_id



from dino_seedwork_be.domain.exceptions import (DomainIllegalArgumentException,
                                                DomainIllegalStateException)
from dino_seedwork_be.logic.assertion_concern import AssertionConcern

# __all__ = ["DomainAssertionConcern"]


class DomainAssertionConcern(AssertionConcern):
    IllegalArgExceptionCls = DomainIllegalArgumentException
    IllegalStExceptionCls = DomainIllegalStateException



from abc import abstractmethod
from typing import Any, Optional, Union

from returns.future import FutureResult
from returns.maybe import Maybe


class AbstractKeyValueRepository:
    _prefix: str

    def __init__(self, prefix: str) -> None:
        self._prefix = prefix

    def prefix(self):
        return self._prefix

    @abstractmethod
    def set(
        self,
        key: str,
        value: Union[bytes, memoryview, str, int, float],
        expired_seconds: Optional[int] = None,
    ) -> FutureResult:
        pass

    @abstractmethod
    def get(self, key: str) -> FutureResult[Maybe, Any]:
        pass



from abc import ABC, abstractmethod
from contextlib import AsyncContextDecorator
from typing import Any, Callable, Generic, List

from dino_seedwork_be.adapters.persistance.sql.DBSessionUser import (
    DBSessionUser, SessionType)


class AbstractUnitOfWork(ABC, Generic[SessionType], AsyncContextDecorator):
    _session_users: List[DBSessionUser[SessionType]]
    _session_factory: Callable[[], SessionType]

    def __init__(
        self,
        session_users: List[DBSessionUser[SessionType]],
        session_factory: Callable[[], SessionType],
    ):
        self._session_users = session_users
        self._session_factory = session_factory
        super().__init__()

    def session_users(self) -> List[DBSessionUser[SessionType]]:
        return self._session_users

    def session_factory(self) -> Callable[[], SessionType]:
        return self._session_factory

    async def __aenter__(self):
        return self

    async def aenter(self):
        return self.__aenter__()

    async def __aexit__(self, *args):
        if args[0] is not None:
            await self.rollback()
        else:
            await self.commit()

    async def aexit(self, *args):
        return self.__aexit__(*args)

    @abstractmethod
    def session(self) -> Any:
        pass

    async def commit(self):
        await self._commit()

    @abstractmethod
    async def _commit(self):
        raise NotImplementedError

    @abstractmethod
    async def rollback(self):
        raise NotImplementedError

    @abstractmethod
    def absord(self):
        raise NotImplementedError



import json

from firebase_admin import App, messaging
from returns.future import FutureFailure, FutureResult, FutureSuccess

from dino_seedwork_be.adapters.messaging.client_notification.AbstractUserNotificationPushDrivingAdapter import \
    UserNotificationPushDrivingAdapter

__all__ = ["FirebaseNotificationDrivingAdapter"]


class FirebaseNotificationDrivingAdapter(UserNotificationPushDrivingAdapter):
    _app: App

    def __init__(self, app: App) -> None:
        self._app = app
        super().__init__()

    def push_to_device(
        self, content: dict, type: str, device_token: str
    ) -> FutureResult:
        try:
            message = messaging.Message(
                data={"type": type, "content": json.dumps(content, default=str)},
                token=device_token,
            )

            response = messaging.send(message)

            return FutureSuccess(response)
        except Exception as error:
            return FutureFailure(error)



from datetime import datetime
from typing import Callable

from dino_seedwork_be.domain.DomainEvent import DomainEvent
from dino_seedwork_be.utils.date import now_utc

from .ProcessId import ProcessId

timeout_factory_type = Callable[[str, ProcessId, int, int, datetime], DomainEvent]


def factory_timeout_event(
    type: str,
    a_process_id: ProcessId,
    retry_count: int = 0,
    a_total_retries_permitted: int = 0,
    occurred_on: datetime = now_utc(),
):
    return DomainEvent(
        type,
        occurred_on=occurred_on,
        props={
            "a_process_id": a_process_id.id(),
            "retry_count": retry_count,
            "total_retries_permitted": a_total_retries_permitted,
        },
    )



from .AlchemyApplicationLifeCycle import (AlchemyApplicationLifeCycle,
                                          DBSessionDomainEventSubs)

__all__ = ["AlchemyApplicationLifeCycle", "DBSessionDomainEventSubs"]



from abc import ABC, abstractmethod
from typing import Any, List

from returns.future import FutureResult

from dino_seedwork_be.domain.DomainEvent import DomainEvent

__all__ = ["EventStore"]


class EventStore(ABC):
    @abstractmethod
    def all_stored_events_since(
        self, a_stored_event_id: int
    ) -> FutureResult[List[DomainEvent], Exception]:
        ...

    @abstractmethod
    def all_stored_events_between(
        self, a_low_stored_event_id: int, a_high_stored_event_id: int
    ) -> FutureResult[List[DomainEvent], Any]:
        ...

    @abstractmethod
    def append(
        self, an_domain_event: DomainEvent
    ) -> FutureResult[DomainEvent, Exception]:
        ...

    @abstractmethod
    def close(self):
        ...

    @abstractmethod
    def count_events(self) -> FutureResult[int, Any]:
        ...



from datetime import datetime, timezone
from typing import Optional


def to_iso_format(aDateTime: Optional[datetime]):
    return aDateTime.isoformat() if aDateTime is not None else None


def now_utc():
    return datetime.now(timezone.utc)



from typing import Generic, List, TypeVar

from dino_seedwork_be.utils.functional import for_each

SessionType = TypeVar("SessionType")


class SessionUserAlreadyHaveSession(Exception):
    ...


class DBSessionUser(Generic[SessionType]):
    _session: SessionType
    _session_preserved = False

    def session(self) -> SessionType:
        return self._session

    def set_session_preserved(self, aBool: bool):
        self._session_preserved = aBool

    def _set_ession(self, session: SessionType):
        self._session = session

    def set_session(self, session: SessionType):
        try:
            if self.session() is not None and not self.is_current_session_closed():
                if not self._session_preserved:
                    raise SessionUserAlreadyHaveSession(
                        f" {str(self)} already ocuppied by a session"
                    )
        except AttributeError:
            self._set_ession(session)
        self._set_ession(session)

    def is_current_session_closed(self) -> bool:
        session = self.session()
        return not (session.new or session.dirty or session.deleted)


class SuperDBSessionUser(Generic[SessionType], DBSessionUser[SessionType]):
    _session_users: List[DBSessionUser] = []
    _session: SessionType

    def set_session(self, session: SessionType):
        self._session = session
        if self.session_users() is not None:
            for_each(
                lambda sessionUser, _: sessionUser.set_session(session),
                self.session_users(),
            )

    def session(self):
        return self._session

    def session_users(self) -> List[DBSessionUser]:
        return self._session_users

    def set_session_users(self, session_users: List[DBSessionUser]):
        self._session_users = session_users



from abc import ABC, abstractmethod
from typing import Any, List

from returns.future import FutureResult

from dino_seedwork_be.domain import DomainEvent

# __all__ = ["DomainEventSubscriber"]


class DomainEventSubscriber(ABC):
    @abstractmethod
    def handle_event(self, an_event: DomainEvent) -> FutureResult[Any, Exception]:
        ...

    @abstractmethod
    def event_type_subscribed(self) -> List[str] | str:
        ...



from returns.maybe import Some
from returns.result import safe

from dino_seedwork_be.logic import AssertionConcern

# __all__ = ["OrderItemMixin"]


class OrderItemMixin(AssertionConcern):
    _order: int = 0

    def get_order(self) -> int:
        return self._order

    @safe
    def set_order(self, order: int):
        self.assert_argument_larger_than(
            order, -1, Some("Order need to a positive number or 0")
        ).unwrap()
        self._order = order
        return "OK"



from typing import List

from returns.future import FutureResult

from ..DomainEvent import DomainEvent
from ..DomainEventSubscriber import DomainEventSubscriber
from .EventStore import EventStore


class EventStoreSubscriber(DomainEventSubscriber):
    _event_store: EventStore

    def __init__(self, event_store: EventStore) -> None:
        self._event_store = event_store
        super().__init__()

    def handle_event(
        self, an_event: DomainEvent
    ) -> FutureResult[DomainEvent, Exception]:
        return self._event_store.append(an_event)

    def event_type_subscribed(self) -> List[str] | str:
        return "*"



from sqlalchemy.ext.asyncio import (AsyncEngine, AsyncSession,
                                    create_async_engine)
from sqlalchemy.orm import sessionmaker
from sqlalchemy.pool import NullPool

__all__ = ["engine_factory", "alchemy_session_factory"]


def engine_factory(uri: str, pool_size: int = 20, max_overflow: int = 20):
    engine = create_async_engine(
        uri,
        isolation_level="READ COMMITTED",
        echo=True,
        future=True,
        # pool_size=pool_size,
        # max_overflow=max_overflow,
        poolclass=NullPool,
    )
    return engine


def alchemy_session_factory(engine: AsyncEngine) -> AsyncSession:
    return sessionmaker(bind=engine, class_=AsyncSession)()



from typing import List

__all__ = ["remove_none", "unique", "shallow_compare_list"]


def remove_none(aList: List) -> List:
    return list(filter(lambda a: a is not None, aList))


def unique(a_list: List) -> List:
    return list(set(a_list))


def shallow_compare_list(a1: List, a2: List):
    return not any((item not in a2) for item in a1)



from abc import ABC, abstractmethod
from types import FunctionType

# __all__ = ["Validator"]

# Will be used as an abstract class for Whole Entity Validating Class
# it will be instantiated inside validate method of entity


class Validator(ABC):

    notification_handler: FunctionType

    def __init__(self, notificationHandler: FunctionType):
        self.set_nofitication_handler(notificationHandler)

    def set_nofitication_handler(self, handler: FunctionType):
        self.notification_handler = handler

    @abstractmethod
    def validate(self):
        pass



from abc import ABC, abstractmethod
from typing import Any

from returns.future import FutureResult
from returns.maybe import Maybe
from returns.result import Result

__all__ = ["NotificationPublisher"]


class NotificationPublisher(ABC):
    @abstractmethod
    def publish_notifications(self) -> FutureResult[Maybe[int], Any]:
        """
        publish unpublished event in event store
        :return: return the last event int id that published already
        """
        pass

    @abstractmethod
    def run(self):
        pass

    @abstractmethod
    def is_ready(self) -> Result[bool, Any]:
        pass

    @abstractmethod
    def get_last_published_notification_id(self) -> FutureResult[Maybe[int], Any]:
        ...

    # @abstractmethod
    # def internal_only_test_confirmation():
    #     pass



from abc import abstractmethod
from typing import Callable, List, Type

from returns.future import FutureResult, FutureSuccess, future_safe
from returns.pipeline import flow
from sqlalchemy.ext.asyncio.session import AsyncSession

from dino_seedwork_be.adapters.logger.SimpleLogger import SIMPLE_LOGGER
from dino_seedwork_be.adapters.persistance.sql.DBSessionUser import (
    DBSessionUser, SuperDBSessionUser)
from dino_seedwork_be.application.AbstractApplicationLifeCycle import \
    AbstractApplicationServiceLifeCycle
from dino_seedwork_be.application.ApplicationLifeCycleUseCase import \
    ApplicationLifeCycleUsecase
from dino_seedwork_be.domain import DomainEventSubscriber
from dino_seedwork_be.domain.DomainEventPublisher import DomainEventPublisher
from dino_seedwork_be.domain.event.EventStoreSubscriber import \
    EventStoreSubscriber
from dino_seedwork_be.implementation.event.sqlalchemy.SqlAlchemyEventStore import (
    SqlAlchemyBaseEvent, SqlAlchemyEventStore)


class DBSessionDomainEventSubs(DomainEventSubscriber, SuperDBSessionUser[AsyncSession]):
    pass


class AlchemyApplicationLifeCycle(AbstractApplicationServiceLifeCycle):
    _domain_event_subscribers: List[Callable[[], DBSessionDomainEventSubs]] = []

    @abstractmethod
    @classmethod
    def session_factory(cls):
        pass

    @classmethod
    def initialize(cls, event_store_model: Type[SqlAlchemyBaseEvent]):
        session = cls.session_factory()
        cls._event_store = SqlAlchemyEventStore(event_store_model)
        cls._event_store.set_session(session)
        return super().initialize()

    @classmethod
    def start_db(
        cls,
        correlation_id: str,
        db_session_users: List[DBSessionUser],
        usecase: ApplicationLifeCycleUsecase,
    ) -> FutureResult:
        session = cls.add_session(correlation_id)[1]
        usecase.set_session(session)
        [session_user.set_session(session) for session_user in db_session_users]
        return FutureSuccess(None)

    @classmethod
    def rollback_db(cls, correlation_id: str) -> FutureResult:
        SIMPLE_LOGGER.warning("[rollback_db] deliberately rollback DB")
        return flow(
            correlation_id,
            cls.get_session_by_correlation,
            lambda session: future_safe(session.rollback)(),
        )

    @classmethod
    def commit_db(cls, correlation_id: str) -> FutureResult:
        return flow(
            correlation_id,
            cls.get_session_by_correlation,
            lambda session: future_safe(session.commit)(),
        )

    @classmethod
    def close_db(cls, correlation_id: str) -> FutureResult:
        return flow(
            correlation_id,
            cls.get_session_by_correlation,
            lambda session: future_safe(session.close()),
        )

    @classmethod
    def set_event_subscribers(
        cls, subscribers: List[Callable[[], DBSessionDomainEventSubs]]
    ):
        cls._domain_event_subscribers = subscribers

    @classmethod
    def event_listen(cls, correlation_id: str) -> FutureResult:
        DomainEventPublisher.instance().reset()
        event_store = cls.event_store()

        DomainEventPublisher.instance().subscribe(EventStoreSubscriber(event_store))
        for publisher_factory in cls._domain_event_subscribers:
            publisher = publisher_factory()
            publisher.set_session(cls.get_session_by_correlation(correlation_id))
            DomainEventPublisher.instance().subscribe(publisher)

        return FutureSuccess("OK")



import logging
from traceback import print_exception

from .AbstractLogger import AbstractLogger

LOG_FORMAT = (
    "%(levelname) -10s %(asctime)s %(name) -30s %(funcName) "
    "-35s %(lineno) -5d: %(message)s"
)

SIMPLE_LOGGER = logging.getLogger(__name__)


class DomainLogger(AbstractLogger):
    def info(self, message: str):
        SIMPLE_LOGGER.info(f"[{self.tag()}] {message}")

    def error(self, message: str, error: Exception | None = None):
        SIMPLE_LOGGER.error(f"[{self.tag()}] {message} f{print_exception(error)}")

    def exception(self, error: Exception | None = None):
        SIMPLE_LOGGER.error(f"[{self.tag()}] f{print_exception(error)}")

    def warning(self, message: str):
        SIMPLE_LOGGER.warning(f"[{self.tag()}] {message}")



from typing import BinaryIO

from dino_seedwork_be.domain.value_object.AbstractValueObject import \
    ValueObject
from dino_seedwork_be.utils.image import (get_image_dimension,
                                          get_image_file_size)

# __all__ = ["File"]


class File(ValueObject):
    _file: BinaryIO
    _content_type: str
    _name: str

    def __init__(self, file: BinaryIO, contentType: str, name: str):
        self._file = file
        self._content_type = contentType
        self._name = name
        super().__init__()

    def dimension(self):
        return get_image_dimension(self.file())

    def size(self) -> float:
        return get_image_file_size(self.file())

    def name(self) -> str:
        return self._name

    def contentType(self):
        return self._content_type

    def file(self):
        return self._file



from .fastapi.basic_handle_exception import (basic_handle_exception,
                                             decor_basic_handler_exception,
                                             safe_basic_handle_exception)
from .fastapi.filters import Filters
from .fastapi.order import Orders
from .utils import (Filter, FilterElement, FilterSet, FilterWithPag,
                    OrderParam, ParamOperator, ParamOperators,
                    ParamWithComparing, PlainFilterSet, RestPayload,
                    error_detail_with_code, error_detail_with_code_standard,
                    error_validation, error_validation_standard,
                    field_error_validation, pagination,
                    plain_order_to_param_order, to_param_orders,
                    to_rest_payload)

__all__ = [
    "basic_handle_exception",
    "decor_basic_handler_exception",
    "safe_basic_handle_exception",
    "Filters",
    "Orders",
    "RestPayload",
    "to_rest_payload",
    "pagination",
    "field_error_validation",
    "error_validation",
    "error_validation_standard",
    "error_detail_with_code_standard",
    "error_detail_with_code",
    "FilterWithPag",
    "ParamOperator",
    "ParamOperators",
    "ParamWithComparing",
    "FilterElement",
    "PlainFilterSet",
    "FilterSet",
    "Filter",
    "OrderParam",
    "plain_order_to_param_order",
    "to_param_orders",
]



import traceback
from typing import Any, List, Optional, Set

from multimethod import multimethod
from pika import BasicProperties
from pika.amqp_object import Method
from pika.channel import Channel
from returns.curry import partial
from returns.functions import tap
from returns.future import FutureResult, FutureSuccess
from returns.pipeline import flow
from returns.pointfree import alt, bind
from returns.result import Failure, Result, Success, safe

from dino_seedwork_be.adapters.logger.SimpleLogger import DomainLogger
from dino_seedwork_be.exceptions import MainException
from dino_seedwork_be.utils.functional import (apply, async_execute,
                                               feed_kwargs,
                                               print_result_with_text,
                                               return_v, tap_excute_future)

from .exceptions import MessageException
from .MessageListener import MessageListener
from .Queue import Queue

# __all__ = ["MessageConsumer"]


class MessageConsumer:
    _auto_acknowledged: bool

    # My closed property, which indicates I have been closed.
    _closed: bool = False

    _consuming: bool = False

    _is_ready: bool = False

    _is_retry: bool = False

    # My messageTypes, which indicates the messages of types I accept.
    _message_types: Set[str] = set([])

    _prefetch_count: int = 1

    # My queue, which is where my messages come from.
    _queue: Queue

    _label: str

    # My tag, which is produced by the broker.
    _tag: str

    logger: DomainLogger

    def __init__(
        self,
        a_queue: Queue,
        is_auto_acknowledged: bool,
        is_retry: bool = False,
        label: str = "",
    ) -> None:
        super().__init__()
        self.set_queue(a_queue)
        self.set_auto_acknowledged(is_auto_acknowledged)
        self.set_message_types(set([]))
        self._is_retry = is_retry
        self._label = label
        self.domain_logger = DomainLogger(self.label())

    @multimethod
    @staticmethod
    def factory(a_queue: Queue) -> Result["MessageConsumer", Any]:
        return Success(MessageConsumer(a_queue, False)).bind(
            lambda consumer: consumer.equalize_message_distribution().map(
                return_v(consumer)
            )
        )

    @staticmethod
    @factory.register
    def _(is_auto_acknowledged: bool, a_queue: Queue) -> Result["MessageConsumer", Any]:
        return Success(MessageConsumer(a_queue, is_auto_acknowledged)).bind(
            lambda consumer: consumer.equalize_message_distribution().map(
                return_v(consumer)
            )
        )

    @staticmethod
    @factory.register
    def _(
        is_auto_acknowledged: bool,
        a_queue: Queue,
        is_retry: bool,
        label: str = "",
    ) -> Result["MessageConsumer", Any]:
        return Success(
            MessageConsumer(a_queue, is_auto_acknowledged, is_retry, label)
        ).bind(
            lambda consumer: consumer.equalize_message_distribution().map(
                return_v(consumer)
            )
        )

    def is_auto_acknowledged(self) -> bool:
        return self._auto_acknowledged

    def set_auto_acknowledged(self, a_bool: bool):
        self._auto_acknowledged = a_bool

    def queue(self) -> Queue:
        return self._queue

    def set_queue(self, a_queue: Queue):
        self._queue = a_queue

    def tag(self) -> str:
        return self._tag

    def label(self) -> str:
        return self._label

    def is_consuming(self) -> bool:
        return self._consuming

    def is_ready(self) -> bool:
        return self._is_ready

    def is_retry(self) -> bool:
        return self._is_retry

    def set_is_ready(self, a_bool: bool):
        self._is_ready = a_bool

    def set_is_consuming(self, a_bool: bool):
        self._consuming = a_bool

    def set_tag(self, a_tag: str):
        self._tag = a_tag

    def message_types(self) -> Set[str]:
        return self._message_types

    def set_message_types(self, a_message_types: Set[str]):
        self._message_types = a_message_types

    def is_closed(self):
        return self._closed

    def equalize_message_distribution(self) -> Result[None, MessageException]:
        """
        Ensure an equalization of message distribution
        across all consumers of this queue.
        """
        try:
            self.queue().channel().unwrap().basic_qos(
                prefetch_count=self._prefetch_count, callback=self.on_basic_qos_ok
            )
            return Success(None)
        except Exception:
            return Failure(MessageException(code="MESSAGE_EQUALIZE_DIS_FAILED"))

    def on_basic_qos_ok(self, _):
        self.domain_logger.info(f"QOS set to: {self._prefetch_count}")
        self.set_is_ready(True)

    def receive_all(self, a_message_listener: MessageListener) -> Result:
        return self.receive_for(a_message_listener)

    def receive_only(
        self, a_message_types: List[str], a_message_listener: MessageListener
    ) -> Result:
        self.set_message_types(set(a_message_types))
        return self.receive_for(a_message_listener)

    def is_target_message_type(self, message_type: Optional[str]) -> bool:
        match len(self.message_types()):
            case 0:
                return True
            case _:
                is_filterred_out = (
                    message_type is None or message_type not in self.message_types()
                )
                return not is_filterred_out

    def receive_for(self, a_message_listener: MessageListener) -> Result:
        queue = self.queue()
        channel = queue.channel().unwrap()

        def ack(
            channel: Channel,
            delivery_tag: int,
        ):
            try:
                match self.is_auto_acknowledged():
                    case False:
                        channel.basic_ack(delivery_tag, False)
                        self.domain_logger.info(
                            f"ACK handle message success {self.message_types()}",
                        )
                return FutureSuccess(None)
            except Exception as error:
                self.domain_logger.error(f"Exception on ACK {error}")
                raise error

        def nak(channel: Channel, delivery_tag: int, is_retry: bool):
            try:
                match self.is_auto_acknowledged():
                    case False:
                        self.domain_logger.info(
                            f"NonACK handle message failed, would retry ? {is_retry}"
                        )
                        channel.basic_nack(delivery_tag, False, is_retry)
            except Exception as error:
                raise error

        def handle_delivery_exception(
            channel: Channel, delivery_tag: int, is_retry: bool, exception: Exception
        ):
            self.domain_logger.info(f"Exception on handle delivery {exception}")
            traceback.print_exc()
            match exception:
                case MessageException():
                    nak(channel, delivery_tag, is_retry)
                case _:
                    nak(channel, delivery_tag, False)
            # do not shutdown
            # raise exception

        def handle_delivery(
            channel: Channel, method: Method, properties: BasicProperties, body: bytes
        ) -> FutureResult:
            match self.is_target_message_type(properties.type):
                case True:
                    self.domain_logger.info(f"Handle delivery {properties.type}")
                    return flow(
                        {
                            "a_type": properties.type,
                            "a_message_id": properties.message_id,
                            "a_time_stamp": properties.timestamp,
                            "a_binary_message": body,
                            "a_delivery_tag": getattr(method, "delivery_tag", 0),
                            "is_redelivery": getattr(method, "delivered", False),
                        },
                        feed_kwargs(a_message_listener.handle_message),
                        bind(
                            tap_excute_future(
                                apply(ack, channel, getattr(method, "delivery_tag", 0))
                            )
                        ),
                        alt(tap(print_result_with_text("error here"))),
                        alt(
                            tap(
                                partial(
                                    handle_delivery_exception,
                                    channel,
                                    getattr(method, "delivery_tag", 0),
                                    self.is_retry(),
                                )
                            )
                        ),
                    )
                case False:
                    return ack(channel, getattr(method, "delivery_tag", 0)).bind(
                        return_v(FutureSuccess("NOT_TARGET_MESSAGE"))
                    )

        try:
            tag = channel.basic_consume(
                queue.name(),
                auto_ack=self.is_auto_acknowledged(),
                on_message_callback=async_execute(handle_delivery),
            )
            channel.add_on_cancel_callback(self.on_consumer_cancelled)
            self.domain_logger.info("Register message listener success")
            self.set_tag(tag)
            self.set_is_consuming(True)
            return Success(None)
        except Exception:
            return Failure(MainException(code="INITIATE_CONSUMER_FAILED"))

    def on_consumer_cancelled(self, method_frame):
        """Invoked by pika when RabbitMQ sends a Basic.Cancel for a consumer
        receiving messages.
        :param pika.frame.Method method_frame: The Basic.Cancel frame
        """
        self.domain_logger.info(
            f"Consumer was cancelled remotely, shutting down: {method_frame}"
        )
        self.close_channel()

    @safe
    def close_channel(self):
        self.queue().channel().map(lambda channel: channel.close())

    def _stop_consuming(self):
        def _cb():
            self.queue.close()

        self.queue().channel().map(
            lambda channel: channel.basic_cancel(self.tag(), _cb)
        )

    @safe
    def close(self):
        """Cleanly shutdown the connection to RabbitMQ by stopping the consumer
        with RabbitMQ. When RabbitMQ confirms the cancellation, on_cancelok
        will be invoked by pika, which will then closing the channel and
        connection. The IOLoop is started again because this method is invoked
        when CTRL-C is pressed raising a KeyboardInterrupt exception. This
        exception stops the IOLoop which needs to be running for pika to
        communicate with RabbitMQ. All of the commands issued prior to starting
        the IOLoop will be buffered but not processed.
        """
        self._closed = True
        match self.is_consuming():
            case True:
                self._stop_consuming()
            case False:
                self.queue().close()

    def run(self):
        self.queue().run()



from datetime import datetime

from dino_seedwork_be.domain.DomainEvent import DomainEvent
from dino_seedwork_be.utils.date import now_utc

from .ProcessId import ProcessId


class ProcessTimedOut(DomainEvent):
    _event_verion: int
    _process_id: ProcessId
    _retry_count: int
    _total_retries_permitted: int

    def __init__(
        self,
        type: str,
        a_process_id: ProcessId,
        retry_count: int = 0,
        a_total_retries_permitted: int = 0,
        occurred_on: datetime = now_utc(),
    ):
        super().__init__(name=type, occurred_on=occurred_on)
        self._process_id = a_process_id
        self._retry_count = retry_count
        self._event_verion = 1
        self._total_retries_permitted = a_total_retries_permitted

    @staticmethod
    def factory(
        type: str,
        a_process_id: ProcessId,
        a_retry_count: int = 0,
        a_total_retries_permitted: int = 0,
        occurred_on: datetime = now_utc(),
    ):
        return ProcessTimedOut(
            type, a_process_id, a_retry_count, a_total_retries_permitted, occurred_on
        )

    def allow_retries(self) -> bool:
        return self.total_retries_permitted() > 0

    def total_retries_permitted(self) -> int:
        return self._total_retries_permitted

    def has_fully_timedout(self):
        return not self.allow_retries() or self.total_retries_reached()

    def total_retries_reached(self):
        return self.retry_count() >= self.total_retries_permitted()

    def retry_count(self) -> int:
        return self._retry_count



import re
import traceback
from datetime import datetime
from decimal import Decimal
from typing import Optional

import jsonpickle
import toolz

from dino_seedwork_be.exceptions import IllegalArgumentException


class JSONReader:
    def deserialize(self, aJson: str):
        try:
            object = jsonpickle.decode(aJson)
            return object
        except Exception:
            traceback.print_exc()
            raise RuntimeError


class AbstractJSONMediaReader:
    _representation: dict = {}
    _json_reader: JSONReader

    def __init__(self, a_json: str):
        self.initialize(a_json)

    def json_reader(self):
        return self._json_reader

    def set_representation(self, a_representation: dict):
        self._representation = a_representation

    def get_representation(self):
        return self._representation

    def initialize(self, a_json: str):
        self._json_reader = JSONReader()
        self.set_representation(self.json_reader().deserialize(a_json))

    def get_value(self, path: str, a_dict: Optional[dict] = None):
        if re.match(r"(^(?:\/[a-zA-Z0-9_]+)+$)", path) is None:
            raise IllegalArgumentException(f"Json Path Reader is illegal {path}")
        source = self.get_representation()
        match a_dict is not None:
            case True:
                source = a_dict
        return toolz.dicttoolz.get_in(path.split("/")[1:], source)

    def string_value(self, path: str, a_dict: Optional[dict] = None) -> str | None:
        value = self.get_value(a_dict=a_dict, path=path)
        if value is None:
            return None
        return str(value)

    def big_decimal_value(
        self, path: str, a_dict: Optional[dict] = None
    ) -> Decimal | None:
        strValue = self.string_value(path, a_dict=a_dict)
        if strValue is None:
            return None
        return Decimal(strValue)

    def boolean_value(self, path: str, a_dict: Optional[dict] = None) -> bool | None:
        str_value = self.string_value(path, a_dict=a_dict)
        if str_value is None:
            return None
        match str_value:
            case "True" | "true":
                return True
            case "False" | "false":
                return False
            case _:
                return bool(str_value)

    def datetime_value(
        self, path: str, a_dict: Optional[dict] = None
    ) -> datetime | None:
        str_value = self.string_value(path, a_dict=a_dict)
        if str_value is None:
            return None
        return datetime.fromisoformat(str_value)

    def float_value(self, path: str, a_dict: Optional[dict] = None) -> float | None:
        str_value = self.string_value(path, a_dict=a_dict)
        if str_value is None:
            return None
        return float(str_value)

    def int_value(self, path: str, a_dict: Optional[dict] = None) -> int | None:
        str_value = self.string_value(path, a_dict=a_dict)
        if str_value is None:
            return None
        return int(float(str_value))



from datetime import datetime, timedelta
from typing import Any

from multimethod import multimethod
from returns.functions import tap
from returns.future import FutureResult, FutureSuccess
from returns.maybe import Some
from returns.pipeline import flow
from returns.pointfree import bind, map_
from returns.result import Result, Success

from dino_seedwork_be.domain import DomainEvent, DomainEventPublisher
from dino_seedwork_be.logic import AssertionConcern
from dino_seedwork_be.utils import (apply, feed_kwargs, now_utc, return_v,
                                    set_protected_attr)

from .ProcessId import ProcessId
from .timeout_event_factory import timeout_factory_type


class TimeConstrainedProcessTracker(AssertionConcern):
    _allowable_duration: int
    _completed: bool = False
    _concurrency_version: int = 0
    _description: str
    _process_id: ProcessId
    _process_informed_of_timeout: bool
    _process_timeout_event_type: str
    _retry_count: int = 0
    _timeout_occurs_on: datetime
    _total_retries_permitted: int

    @multimethod
    def __init__(self) -> None:
        super().__init__()

    def __eq__(self, other):
        if isinstance(other, TimeConstrainedProcessTracker):
            return other.process_id().id() == self.process_id().id()
        return False

    @multimethod
    @__init__.register
    def _(
        self,
        a_process_id: ProcessId,
        timeout_occurs_on: datetime,
        an_allowable_duration: int,
        a_retry_count: int,
        a_total_retries_permitted: int,
        a_process_timedout_event_type: str,
        is_process_informed_of_timeout: bool,
        completed: bool,
        a_description: str = "",
    ):
        flow(
            a_process_id,
            self.set_process_id,
            bind(apply(self.set_description, a_description)),
            bind(apply(self.set_timeout_occurs_on, timeout_occurs_on)),
            bind(apply(self.set_allowable_duration, an_allowable_duration)),
            bind(apply(self.set_total_retries_permitted, a_total_retries_permitted)),
            bind(apply(self.set_retry_count, a_retry_count)),
            bind(
                apply(
                    self.set_process_timeout_event_type, a_process_timedout_event_type
                )
            ),
            bind(
                apply(
                    self.set_process_informed_of_timeout, is_process_informed_of_timeout
                )
            ),
            bind(apply(self.set_completed, completed)),
        ).unwrap()

    @staticmethod
    def factory(
        a_process_id: ProcessId,
        an_original_start_time: datetime,
        an_allowable_duration: int,
        a_total_retries_permitted: int,
        a_process_timedout_event_type: str,
        a_description: str = "",
    ) -> Result["TimeConstrainedProcessTracker", Any]:
        a_new_tracker = TimeConstrainedProcessTracker()
        return flow(
            None,
            lambda _: a_new_tracker.set_process_id(a_process_id),
            bind(lambda _: a_new_tracker.set_description(a_description)),
            bind(
                lambda _: a_new_tracker.set_timeout_occurs_on(
                    an_original_start_time + timedelta(seconds=an_allowable_duration)
                )
            ),
            bind(
                lambda _: a_new_tracker.set_total_retries_permitted(
                    a_total_retries_permitted
                )
            ),
            bind(
                lambda _: a_new_tracker.set_process_timeout_event_type(
                    a_process_timedout_event_type
                )
            ),
            bind(lambda _: a_new_tracker.set_allowable_duration(an_allowable_duration)),
            map_(return_v(a_new_tracker)),
        )

    def allowable_duration(self) -> int:
        return self._allowable_duration

    def completed(self) -> bool:
        return self._completed

    def set_completed(self, a_bool: bool) -> Result:
        self._completed = a_bool
        return Success(None)

    def description(self) -> str:
        return self._description

    def set_description(self, a_str: str) -> Result:
        self._description = a_str
        return Success(None)

    def process_id(self) -> ProcessId:
        return self._process_id

    def set_process_id(self, a_process_id) -> Result:
        self._process_id = a_process_id
        return Success(None)

    def is_process_informed_of_timeout(self) -> bool:
        return self._process_informed_of_timeout

    def process_timeout_event_type(self) -> str:
        return self._process_timeout_event_type

    def set_process_timeout_event_type(self, a_type: str) -> Result:
        self._process_timeout_event_type = a_type
        return Success(None)

    def timeout_occurs_on(self) -> datetime:
        return self._timeout_occurs_on

    def has_timed_out(self) -> bool:
        return now_utc() >= self.timeout_occurs_on()

    def total_retries_permitted(self) -> int:
        return self._total_retries_permitted

    def set_total_retries_permitted(self, a_num: int) -> Result:
        self._total_retries_permitted = a_num
        return Success(None)

    def total_retries_reached(self):
        return self.retry_count() > self.total_retries_permitted()

    def retry_count(self) -> int:
        return self._retry_count

    def set_retry_count(self, a_num: int) -> Result:
        return flow(
            self.assert_argument_range(
                a_value=a_num,
                a_minium=0,
                a_maximum=self.total_retries_permitted() + 1,
                code=Some("RETRY_COUNT_VALID_VALUE"),
            ),
            map_(tap(lambda _: set_protected_attr(self, "_retry_count", a_num))),
        )

    def process_timeout_event(
        self, timeout_event_factory: timeout_factory_type
    ) -> DomainEvent:
        return timeout_event_factory(
            self.process_timeout_event_type(),
            self.process_id(),
            self.retry_count(),
            self.total_retries_permitted(),
            now_utc(),
        )

    def set_process_informed_of_timeout(self, a_bool: bool) -> Result:
        self._process_informed_of_timeout = a_bool
        return Success(None)

    def set_timeout_occurs_on(self, a_datetime: datetime) -> Result:
        self._timeout_occurs_on = a_datetime
        return Success(None)

    def set_allowable_duration(self, an_allowable_duration: int) -> Result:
        return self.assert_argument_larger_than(
            an_allowable_duration,
            0,
            code=Some("ALLOW_DURATION_MUST_BE_LARGER_THAN_ZERO"),
        ).map(
            tap(
                lambda _: set_protected_attr(
                    self, "_allowable_duration", an_allowable_duration
                )
            )
        )

    def increment_retry_count(self):
        self._retry_count += 1

    def is_completed(self) -> bool:
        return self._completed

    def inform_process_timedout(
        self, timeout_event_factory: timeout_factory_type
    ) -> FutureResult:
        match not self.is_process_informed_of_timeout() and self.has_timed_out():
            case True:
                match self.total_retries_permitted():
                    case 0:
                        return flow(
                            self.process_timeout_event(timeout_event_factory),
                            map_(DomainEventPublisher.instance().publish),
                            map_(
                                tap(
                                    lambda _: self.set_process_informed_of_timeout(True)
                                )
                            ),
                        )
                    case _:

                        def update_tracker():
                            self.increment_retry_count()
                            match self.total_retries_reached():
                                case True:
                                    self.set_process_informed_of_timeout(True)
                                case False:
                                    self.set_timeout_occurs_on(
                                        now_utc()
                                        + timedelta(seconds=self.allowable_duration())
                                    )

                        update_tracker()

                        return flow(
                            self.process_timeout_event(timeout_event_factory),
                            DomainEventPublisher.instance().publish,
                        )
        return FutureSuccess(None)



from abc import ABC, abstractmethod
from typing import Any, List

from returns.future import FutureResult
from returns.maybe import Maybe

from .ProcessId import ProcessId
from .TimeConstrainedProcessTracker import TimeConstrainedProcessTracker
from .timeout_event_factory import timeout_factory_type


class TimeConstrainedProcessTrackerRepository(ABC):
    @abstractmethod
    def add(
        self, a_time_constrained_process_tracker: TimeConstrainedProcessTracker
    ) -> FutureResult:
        ...

    @abstractmethod
    def all_timed_out() -> FutureResult[List[TimeConstrainedProcessTracker], Any]:
        ...

    @abstractmethod
    def save(
        self, a_time_constrained_process_tracker: TimeConstrainedProcessTracker
    ) -> FutureResult:
        ...

    @abstractmethod
    def tracker_of_process_id(
        self, a_process_id: ProcessId
    ) -> FutureResult[Maybe[TimeConstrainedProcessTracker], Any]:
        ...

    @abstractmethod
    def timeout_event_factory(self) -> timeout_factory_type:
        ...

    @abstractmethod
    def complete_tracker(self, process_id: ProcessId) -> FutureResult:
        pass



import traceback

from fastapi import HTTPException, status
from returns.future import FutureFailure

from dino_seedwork_be.adapters.logger.SimpleLogger import SIMPLE_LOGGER
from dino_seedwork_be.exceptions import MainException

from ..utils import error_detail_with_code


def safe_basic_handle_exception(exception):
    match exception:
        case MainException():
            return FutureFailure(
                HTTPException(
                    status_code=status.HTTP_400_BAD_REQUEST,
                    detail=error_detail_with_code(exception),
                )
            )
        case _:
            SIMPLE_LOGGER.info("Error Occured: %s", exception)
            traceback.print_exc()
            return FutureFailure(
                HTTPException(
                    status_code=status.HTTP_500_INTERNAL_SERVER_ERROR,
                    detail=str(exception),
                )
            )


def basic_handle_exception(exception):
    match exception:
        case MainException():
            raise HTTPException(
                status_code=status.HTTP_400_BAD_REQUEST,
                detail=error_detail_with_code(exception),
            )

        case _:
            SIMPLE_LOGGER.info("Error Occured: %s", exception)
            traceback.print_exc()
            # raise exception
            raise HTTPException(
                status_code=status.HTTP_500_INTERNAL_SERVER_ERROR,
                detail=str(exception),
            )


def decor_basic_handler_exception(func):
    def wrapper(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        except Exception as error:
            basic_handle_exception(error)

    return wrapper



from abc import ABC, abstractmethod
from typing import Any, Generic, TypeVar

from returns.functions import tap
from returns.pipeline import flow
from returns.pointfree import bind, map_
from returns.result import Result

from dino_seedwork_be.utils import get_class_name, set_protected_attr, unwrap

from .AbstractValueObject import ValueObject

IdRawType = TypeVar("IdRawType")

# __all__ = ["AbstractIdentity"]


class AbstractIdentity(ABC, ValueObject, Generic[IdRawType]):

    _id: IdRawType

    def __init__(self, id: IdRawType) -> None:
        super().__init__()
        unwrap(self.set_id(id))

    def id(self) -> IdRawType:
        return self._id

    def __eq__(self, __o: object) -> bool:
        match __o:
            case AbstractIdentity():
                return type(__o) == type(self) and self.id() == __o.id()
            case _:
                return False

    def __str__(self) -> str:
        return f"{get_class_name(self)} [id={str(id)}]"

    def __hash__(self) -> int:
        return hash(self.get_raw())

    @abstractmethod
    def validate(self, an_id: IdRawType) -> Result[IdRawType, Any]:
        pass

    def set_id(self, an_id: IdRawType):
        def set(an_id):
            self._id = an_id

        return flow(
            an_id,
            self.assert_argument_not_null,
            bind(self.validate),
            map_(tap(set)),
        )

    def get_raw(self) -> IdRawType:
        return self.id()

    def get_raw_str(self) -> str:
        return str(self.id())

    def __getstate__(self):
        return self.__str__()



from datetime import datetime
from typing import Any

from returns.maybe import Maybe, Nothing, Some
from returns.result import Result

from dino_seedwork_be.domain.DomainEvent import DomainEvent
from dino_seedwork_be.logic.assertion_concern import AssertionConcern

from .EventSerializer import EventSerializer

# __all__ = ["StoredEvent"]


class StoredEvent(AssertionConcern):
    _body: str
    _id: Maybe[int]
    _occurred_on: datetime
    _type_name: str

    def __init__(
        self,
        body: str,
        occurred_on: datetime,
        type_name: str,
        id: Maybe[int] = Nothing,
    ) -> None:
        self.set_body(body)
        match id:
            case Some(int(v)):
                self.set_id(v)
            case _:
                self._id = Nothing
        self.set_occurred_on(occurred_on)
        self.set_type_name(type_name)

    @staticmethod
    def factory(
        body: str, occurred_on: datetime, type_name: str, id: Maybe[int] = Nothing
    ):
        return StoredEvent(body, occurred_on, type_name, id)

    def set_body(self, an_event_body: str):
        self.assert_argument_not_null(
            an_event_body, a_message=Some("Event body cannot be empty")
        )
        self.assert_argument_range(
            len(an_event_body),
            a_minium=1,
            a_maximum=65000,
            a_message=Some("The event body must be 65000 character or less"),
        )
        self._body = an_event_body

    def set_id(self, an_id: int):
        self._id = Some(an_id)

    def set_occurred_on(self, a_datetime: datetime):
        self._occurred_on = a_datetime

    def set_type_name(self, a_type_name: str):
        self.assert_argument_not_empty(
            Some(a_type_name), Some("The event type name is required")
        )
        self.assert_argument_length(
            a_type_name,
            a_minimum=1,
            a_maximum=100,
            a_message=Some("The event type name must be 100 characters or less"),
        )
        self._type_name = a_type_name

    def id(self) -> Maybe[int]:
        return self._id

    def body(self) -> str:
        return self._body

    def type_name(self) -> str:
        return self._type_name

    def occurred_on(self) -> datetime:
        return self._occurred_on

    def to_domain_event(self) -> Result[DomainEvent, Any]:
        return EventSerializer.instance().deserialize(self.body())
